import MDAnalysis as mda
import numpy as np
import pytest
from pandas import DataFrame
from rdkit.DataStructs import ExplicitBitVect, UIntSparseIntVect

from prolif.datafiles import datapath
from prolif.fingerprint import Fingerprint
from prolif.interactions.base import _INTERACTIONS, Interaction
from prolif.molecule import sdf_supplier
from prolif.residue import ResidueId


class Dummy(Interaction):
    def detect(self, res1, res2):
        yield self.metadata(res1, res2, (2,), (4,), distance=4.2)


def test_interaction_base(sdf_suppl):
    interaction = Dummy()
    _repr = repr(interaction)
    assert _repr.startswith("<") and ".Dummy at " in _repr
    assert callable(interaction)
    mol = sdf_suppl[0]
    metadata = next(interaction.detect(mol, mol))
    assert metadata["indices"] == {"ligand": (2,), "protein": (4,)}
    assert metadata["parent_indices"] == {"ligand": (2,), "protein": (4,)}
    # invert
    inverted = Dummy.invert_role("Dummy", "inverted")
    inverted_interaction = inverted()
    metadata = next(inverted_interaction.detect(mol, mol))
    assert metadata["indices"] == {"ligand": (4,), "protein": (2,)}
    assert metadata["parent_indices"] == {"ligand": (4,), "protein": (2,)}


class TestFingerprint:
    @pytest.fixture(scope="class")
    def fp(self):
        yield Fingerprint()
        _INTERACTIONS.pop("Dummy", None)

    @pytest.fixture(scope="class")
    def fp_count(self):
        yield Fingerprint(count=True)
        _INTERACTIONS.pop("Dummy", None)

    @pytest.fixture(scope="class", params=["fp", "fp_count"])
    def any_fp(self, request):
        return request.getfixturevalue(request.param)

    @pytest.fixture(scope="class")
    def fp_simple(self):
        yield Fingerprint(["Hydrophobic"])
        _INTERACTIONS.pop("Dummy", None)

    def test_init(self, fp_simple):
        assert "Hydrophobic" in fp_simple.interactions.keys()
        assert hasattr(fp_simple, "hydrophobic") and callable(fp_simple.hydrophobic)
        assert "Dummy" not in fp_simple.interactions.keys()
        assert hasattr(fp_simple, "dummy") and callable(fp_simple.dummy)
        assert "_BaseHBond" not in fp_simple.interactions.keys()
        assert not hasattr(fp_simple, "_basehbond")
        assert "Interaction" not in fp_simple.interactions.keys()
        assert not hasattr(fp_simple, "interaction")

    def test_init_all(self):
        fp = Fingerprint("all")
        for name, func in fp.interactions.items():
            assert getattr(fp, name.lower()) is func.__wrapped__

    def test_n_interactions(self, fp):
        assert fp.n_interactions == len(fp.interactions)

    def test_bitvector(self, any_fp, ligand_mol, protein_mol):
        bv = any_fp.bitvector(ligand_mol[0], protein_mol["ASP129.A"])
        assert len(bv) == any_fp.n_interactions
        assert (bv == 0).any()
        assert bv.sum() > 0
        if any_fp.count:
            assert (bv > 1).any()

    def test_metadata(self, any_fp, ligand_mol, protein_mol):
        metadata = any_fp.metadata(ligand_mol[0], protein_mol["ASP129.A"])
        assert metadata
        assert "HBAcceptor" not in metadata
        assert all(
            isinstance(cationic["indices"]["ligand"], tuple)
            for cationic in metadata["Cationic"]
        )

    def test_run_residues(self, fp_simple, u, ligand_ag, protein_ag):
        fp_simple.run(
            u.trajectory[0:1],
            ligand_ag,
            protein_ag,
            residues=["TYR109.A"],
            progress=False,
        )
        lig_id = ResidueId.from_string("LIG1.G")
        assert hasattr(fp_simple, "ifp")
        assert len(fp_simple.ifp) == 1
        res = ResidueId.from_string("TYR109.A")
        assert (lig_id, res) in fp_simple.ifp[0].keys()
        fp_simple.run(
            u.trajectory[1:2],
            ligand_ag,
            protein_ag,
            residues="all",
            progress=False,
        )
        assert hasattr(fp_simple, "ifp")
        assert len(fp_simple.ifp) == 1
        res = ResidueId.from_string("TRP125.A")
        assert (lig_id, res) in fp_simple.ifp[1].keys()
        fp_simple.run(
            u.trajectory[:3], ligand_ag, protein_ag, residues=None, progress=False
        )
        assert hasattr(fp_simple, "ifp")
        assert len(fp_simple.ifp) == 3
        assert len(fp_simple.ifp[0]) > 1
        res = ResidueId.from_string("ALA216.A")
        assert (lig_id, res) in fp_simple.ifp[0].keys()
        u.trajectory[0]

    def test_generate(self, fp_simple, ligand_mol, protein_mol):
        ifp = fp_simple.generate(ligand_mol, protein_mol)
        key = (ResidueId("LIG", 1, "G"), ResidueId("VAL", 201, "A"))
        bv = ifp[key]
        assert isinstance(bv, np.ndarray)
        assert bv[0] is np.True_

    def test_generate_metadata(self, fp_simple, ligand_mol, protein_mol):
        ifp = fp_simple.generate(ligand_mol, protein_mol, metadata=True)
        key = (ResidueId("LIG", 1, "G"), ResidueId("VAL", 201, "A"))
        int_data = ifp[key]
        assert "Hydrophobic" in int_data

    def test_run(self, fp_simple, u, ligand_ag, protein_ag):
        fp_simple.run(
            u.trajectory[0:1], ligand_ag, protein_ag, residues=None, progress=False
        )
        assert hasattr(fp_simple, "ifp")
        ifp = fp_simple.ifp[0]
        interactions = next(iter(ifp.values()))
        assert isinstance(interactions, dict)
        metadata_tuple = next(iter(interactions.values()))
        assert all(
            [
                key in metadata_tuple[0]
                for key in ["indices", "parent_indices", "distance"]
            ]
        )

    def test_run_from_iterable(self, fp_simple, protein_mol):
        path = str(datapath / "vina" / "vina_output.sdf")
        lig_suppl = list(sdf_supplier(path))
        fp_simple.run_from_iterable(lig_suppl[:2], protein_mol, progress=False)
        assert len(fp_simple.ifp) == 2

    def test_to_df(self, fp_simple, u, ligand_ag, protein_ag):
        with pytest.raises(AttributeError, match="use the `run` method"):
            Fingerprint().to_dataframe()
        fp_simple.run(
            u.trajectory[:3], ligand_ag, protein_ag, residues=None, progress=False
        )
        df = fp_simple.to_dataframe()
        assert isinstance(df, DataFrame)
        assert len(df) == 3

    def test_to_df_kwargs(self, fp_simple, u, ligand_ag, protein_ag):
        fp_simple.run(
            u.trajectory[:3], ligand_ag, protein_ag, residues=None, progress=False
        )
        df = fp_simple.to_dataframe(dtype=np.uint8)
        assert df.dtypes[0].type is np.uint8
        df = fp_simple.to_dataframe(drop_empty=False)
        resids = set([key for d in fp_simple.ifp.values() for key in d.keys()])
        assert df.shape == (3, len(resids))

    def test_to_bitvector(self, fp_simple, u, ligand_ag, protein_ag):
        with pytest.raises(AttributeError, match="use the `run` method"):
            Fingerprint().to_bitvectors()
        fp_simple.run(
            u.trajectory[:3], ligand_ag, protein_ag, residues=None, progress=False
        )
        bvs = fp_simple.to_bitvectors()
        assert isinstance(bvs[0], ExplicitBitVect)
        assert len(bvs) == 3

    def test_to_countvectors(self, fp_count, u, ligand_ag, protein_ag):
        with pytest.raises(AttributeError, match="use the `run` method"):
            Fingerprint().to_countvectors()
        fp_count.run(
            u.trajectory[:3], ligand_ag, protein_ag, residues=None, progress=False
        )
        cvs = fp_count.to_countvectors()
        assert isinstance(cvs[0], UIntSparseIntVect)
        assert cvs[0][0] == 4
        assert len(cvs) == 3

    def test_list_avail(self):
        available = Fingerprint.list_available()
        assert "Hydrophobic" in available
        assert "HBDonor" in available
        assert "Distance" not in available
        assert "Interaction" not in available
        available = Fingerprint.list_available(show_hidden=True)
        assert "Hydrophobic" in available
        assert "HBDonor" in available
        assert "Distance" in available
        assert "Interaction" not in available

    def test_unknown_interaction(self):
        with pytest.raises(
            NameError, match=r"Unknown interaction\(s\) in 'interactions': foo"
        ):
            Fingerprint(["Cationic", "foo"])
        with pytest.raises(
            NameError, match=r"Unknown interaction\(s\) in 'parameters': bar"
        ):
            Fingerprint(["Cationic"], parameters={"bar": {}})

    @pytest.fixture
    def fp_to_pickle(self, fp, protein_mol):
        path = str(datapath / "vina" / "vina_output.sdf")
        lig_suppl = list(sdf_supplier(path))
        fp.run_from_iterable(lig_suppl[:2], protein_mol, progress=False)
        return fp

    @pytest.fixture
    def fp_unpkl(self, fp_to_pickle):
        pkl = fp_to_pickle.to_pickle()
        return Fingerprint.from_pickle(pkl)

    @pytest.fixture
    def fp_unpkl_file(self, fp_to_pickle, tmp_path):
        pkl_path = tmp_path / "fp.pkl"
        fp_to_pickle.to_pickle(pkl_path)
        return Fingerprint.from_pickle(pkl_path)

    @pytest.fixture(params=["fp_unpkl", "fp_unpkl_file"])
    def fp_pkled(self, request):
        return request.getfixturevalue(request.param)

    def test_pickle(self, fp_to_pickle, fp_pkled):
        assert fp_to_pickle.interactions.keys() == fp_pkled.interactions.keys()
        assert len(fp_to_pickle.ifp) == len(fp_pkled.ifp)
        for frame_ifp, frame_pkl_ifp in zip(fp_to_pickle.ifp, fp_pkled.ifp):
            assert frame_ifp == frame_pkl_ifp

    def test_pickle_custom_interaction(self, fp_unpkl):
        assert hasattr(fp_unpkl, "dummy")
        assert callable(fp_unpkl.dummy)

    def test_run_multiproc_serial_same(self, fp, u, ligand_ag, protein_ag):
        fp.run(u.trajectory[0:100:10], ligand_ag, protein_ag, n_jobs=1, progress=False)
        serial = fp.to_dataframe()
        fp.run(
            u.trajectory[0:100:10], ligand_ag, protein_ag, n_jobs=None, progress=False
        )
        multi = fp.to_dataframe()
        assert serial.equals(multi)

    def test_run_iter_multiproc_serial_same(self, fp, protein_mol):
        run = fp.run_from_iterable
        path = str(datapath / "vina" / "vina_output.sdf")
        lig_suppl = sdf_supplier(path)
        run(lig_suppl, protein_mol, n_jobs=1, progress=False)
        serial = fp.to_dataframe()
        run(lig_suppl, protein_mol, n_jobs=None, progress=False)
        multi = fp.to_dataframe()
        assert serial.equals(multi)

    def test_converter_kwargs_raises_error(self, fp, u, ligand_ag, protein_ag):
        with pytest.raises(
            ValueError, match="converter_kwargs must be a list of 2 dicts"
        ):
            fp.run(
                u.trajectory[0:5],
                ligand_ag,
                protein_ag,
                n_jobs=1,
                progress=False,
                converter_kwargs=[dict(force=True)],
            )

    @pytest.mark.parametrize("n_jobs", [1, 2])
    def test_converter_kwargs(self, fp, n_jobs):
        u = mda.Universe.from_smiles("O=C=O.O=C=O")
        lig, prot = u.atoms.fragments
        fp.run(
            u.trajectory,
            lig,
            prot,
            n_jobs=n_jobs,
            converter_kwargs=[dict(force=True), dict(force=True)],
        )
        assert fp.ifp

    def test_interaction_params(self):
        fp = Fingerprint()
        assert fp.hydrophobic.distance == 4.5
        fp = Fingerprint(parameters={"Hydrophobic": {"distance": 1.0}})
        assert fp.hydrophobic.distance == 1.0
        fp = Fingerprint()
        assert fp.hydrophobic.distance == 4.5



import builtins

IS_NOTEBOOK = hasattr(builtins, "__IPYTHON__")

separated_interaction_colors = {
    "Hydrophobic": "#59e382",
    "VdWContact": "#dfab43",
    "HBAcceptor": "#59bee3",
    "HBDonor": "#239fcd",
    "XBAcceptor": "#ff9f02",
    "XBDonor": "#ce8000",
    "Cationic": "#e35959",
    "Anionic": "#5979e3",
    "CationPi": "#e359d8",
    "PiCation": "#ea85e2",
    "PiStacking": "#b559e3",
    "EdgeToFace": "#c885ea",
    "FaceToFace": "#a22ddc",
    "MetalAcceptor": "#7da982",
    "MetalDonor": "#609267",
}

grouped_interaction_colors = {
    "Hydrophobic": "#59e382",
    "VdWContact": "#dfab43",
    "HBAcceptor": "#59bee3",
    "HBDonor": "#59bee3",
    "XBAcceptor": "#ff9f02",
    "XBDonor": "#ff9f02",
    "Cationic": "#e35959",
    "Anionic": "#5979e3",
    "CationPi": "#e359d8",
    "PiCation": "#e359d8",
    "PiStacking": "#b559e3",
    "EdgeToFace": "#b559e3",
    "FaceToFace": "#b559e3",
    "MetalAcceptor": "#7da982",
    "MetalDonor": "#7da982",
}



import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis.transformations import rotateby, translate
from rdkit import Chem, RDLogger

import prolif
from prolif.fingerprint import Fingerprint
from prolif.interactions import VdWContact
from prolif.interactions.base import _INTERACTIONS, Interaction, get_mapindex

from . import mol2factory

# disable rdkit warnings
lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)


@pytest.fixture(scope="module")
def benzene():
    benzene = mda.Universe(prolif.datafiles.datapath / "benzene.mol2")
    elements = mda.topology.guessers.guess_types(benzene.atoms.names)
    benzene.add_TopologyAttr("elements", elements)
    benzene.segments.segids = np.array(["U1"], dtype=object)
    benzene.transfer_to_memory()
    return benzene


@pytest.fixture(scope="module")
def interaction_instances():
    return {
        name: cls()
        for name, cls in _INTERACTIONS.items()
        if name != "Interaction" and not name.startswith("_")
    }


@pytest.fixture(scope="module")
def lig_mol(request):
    return getattr(mol2factory, request.param)()


@pytest.fixture(scope="module")
def prot_mol(request):
    return getattr(mol2factory, request.param)()


@pytest.fixture(scope="module")
def interaction_qmol(request, interaction_instances):
    int_name, parameter = request.param.split(".")
    return getattr(interaction_instances[int_name], parameter)


class TestInteractions:
    @pytest.fixture(scope="class")
    def fingerprint(self):
        return Fingerprint()

    @pytest.mark.parametrize(
        "func_name, lig_mol, prot_mol, expected",
        [
            ("cationic", "cation", "anion", True),
            ("cationic", "anion", "cation", False),
            ("cationic", "cation", "benzene", False),
            ("anionic", "cation", "anion", False),
            ("anionic", "anion", "cation", True),
            ("anionic", "anion", "benzene", False),
            ("cationpi", "cation", "benzene", True),
            ("cationpi", "cation_false", "benzene", False),
            ("cationpi", "benzene", "cation", False),
            ("cationpi", "cation", "cation", False),
            ("cationpi", "benzene", "benzene", False),
            ("pication", "benzene", "cation", True),
            ("pication", "benzene", "cation_false", False),
            ("pication", "cation", "benzene", False),
            ("pication", "cation", "cation", False),
            ("pication", "benzene", "benzene", False),
            ("pistacking", "benzene", "etf", True),
            ("pistacking", "etf", "benzene", True),
            ("pistacking", "ftf", "benzene", True),
            ("pistacking", "benzene", "ftf", True),
            ("facetoface", "benzene", "ftf", True),
            ("facetoface", "ftf", "benzene", True),
            ("facetoface", "benzene", "etf", False),
            ("facetoface", "etf", "benzene", False),
            ("edgetoface", "benzene", "etf", True),
            ("edgetoface", "etf", "benzene", True),
            ("edgetoface", "benzene", "ftf", False),
            ("edgetoface", "ftf", "benzene", False),
            ("hydrophobic", "benzene", "etf", True),
            ("hydrophobic", "benzene", "ftf", True),
            ("hydrophobic", "benzene", "chlorine", False),
            ("hydrophobic", "benzene", "bromine", True),
            ("hydrophobic", "benzene", "anion", False),
            ("hydrophobic", "benzene", "cation", False),
            ("hbdonor", "hb_donor", "hb_acceptor", True),
            ("hbdonor", "hb_donor", "hb_acceptor_false", False),
            ("hbdonor", "hb_acceptor", "hb_donor", False),
            ("hbacceptor", "hb_acceptor", "hb_donor", True),
            ("hbacceptor", "hb_acceptor_false", "hb_donor", False),
            ("hbacceptor", "hb_donor", "hb_acceptor", False),
            ("xbdonor", "xb_donor", "xb_acceptor", True),
            ("xbdonor", "xb_donor", "xb_acceptor_false_xar", False),
            ("xbdonor", "xb_donor", "xb_acceptor_false_axd", False),
            ("xbdonor", "xb_acceptor", "xb_donor", False),
            ("xbacceptor", "xb_acceptor", "xb_donor", True),
            ("xbacceptor", "xb_acceptor_false_xar", "xb_donor", False),
            ("xbacceptor", "xb_acceptor_false_axd", "xb_donor", False),
            ("xbacceptor", "xb_donor", "xb_acceptor", False),
            ("metaldonor", "metal", "ligand", True),
            ("metaldonor", "metal_false", "ligand", False),
            ("metaldonor", "ligand", "metal", False),
            ("metalacceptor", "ligand", "metal", True),
            ("metalacceptor", "ligand", "metal_false", False),
            ("metalacceptor", "metal", "ligand", False),
            ("vdwcontact", "benzene", "etf", True),
            ("vdwcontact", "hb_acceptor", "metal_false", False),
        ],
        indirect=["lig_mol", "prot_mol"],
    )
    def test_interaction(self, fingerprint, func_name, lig_mol, prot_mol, expected):
        interaction = getattr(fingerprint, func_name)
        assert next(interaction(lig_mol[0], prot_mol[0]), False) is expected

    def test_warning_supersede(self):
        old = id(_INTERACTIONS["Hydrophobic"])
        with pytest.warns(UserWarning, match="interaction has been superseded"):

            class Hydrophobic(Interaction):
                def detect(self):
                    pass

        new = id(_INTERACTIONS["Hydrophobic"])
        assert old != new
        # fix dummy Hydrophobic class being reused in later unrelated tests

        class Hydrophobic(prolif.interactions.Hydrophobic):
            __doc__ = prolif.interactions.Hydrophobic.__doc__

    def test_error_no_detect(self):
        with pytest.raises(
            TypeError, match="Can't instantiate interaction class _Dummy"
        ):

            class _Dummy(Interaction):
                pass

    @pytest.mark.parametrize("index", [0, 1, 3, 42, 78])
    def test_get_mapindex(self, index, ligand_mol):
        parent_index = get_mapindex(ligand_mol[0], index)
        assert parent_index == index

    def test_vdwcontact_tolerance_error(self):
        with pytest.raises(ValueError, match="`tolerance` must be 0 or positive"):
            VdWContact(tolerance=-1)

    @pytest.mark.parametrize(
        "lig_mol, prot_mol", [("benzene", "cation")], indirect=["lig_mol", "prot_mol"]
    )
    def test_vdwcontact_cache(self, lig_mol, prot_mol):
        vdw = VdWContact()
        assert vdw._vdw_cache == {}
        vdw.detect(lig_mol[0], prot_mol[0])
        for (lig, res), value in vdw._vdw_cache.items():
            vdw_dist = vdw.vdwradii[lig] + vdw.vdwradii[res] + vdw.tolerance
            assert vdw_dist == value

    @pytest.mark.parametrize(
        "lig_mol, prot_mol", [("benzene", "cation")], indirect=["lig_mol", "prot_mol"]
    )
    def test_vdwcontact_vdwradii_update(self, lig_mol, prot_mol):
        vdw = VdWContact(vdwradii={"Na": 0})
        metadata = vdw.detect(lig_mol[0], prot_mol[0])
        assert next(metadata, None) is None

    @pytest.mark.parametrize(
        ["interaction_qmol", "smiles", "expected"],
        [
            ("Hydrophobic.lig_pattern", "C", 1),
            ("Hydrophobic.lig_pattern", "C=[SH2]", 1),
            ("Hydrophobic.lig_pattern", "c1cscc1", 5),
            ("Hydrophobic.lig_pattern", "CSC", 3),
            ("Hydrophobic.lig_pattern", "CS(C)(C)C", 4),
            ("Hydrophobic.lig_pattern", "FC(F)(F)F", 0),
            ("Hydrophobic.lig_pattern", "BrI", 2),
            ("Hydrophobic.lig_pattern", "C=O", 0),
            ("Hydrophobic.lig_pattern", "C=N", 0),
            ("Hydrophobic.lig_pattern", "CF", 0),
            ("HBAcceptor.prot_pattern", "[OH2]", 2),
            ("HBAcceptor.prot_pattern", "[NH3]", 3),
            ("HBAcceptor.prot_pattern", "[NH4+]", 4),
            ("HBAcceptor.prot_pattern", "[SH2]", 2),
            ("HBAcceptor.prot_pattern", "O=C=O", 0),
            ("HBAcceptor.prot_pattern", "c1c[nH+]ccc1", 0),
            ("HBAcceptor.prot_pattern", "c1c[nH]cc1", 1),
            ("HBAcceptor.lig_pattern", "O", 1),
            ("HBAcceptor.lig_pattern", "N", 1),
            ("HBAcceptor.lig_pattern", "[NH4+]", 0),
            ("HBAcceptor.lig_pattern", "N-C=O", 1),
            ("HBAcceptor.lig_pattern", "N-C=[SH2]", 0),
            ("HBAcceptor.lig_pattern", "[nH+]1ccccc1", 0),
            ("HBAcceptor.lig_pattern", "n1ccccc1", 1),
            ("HBAcceptor.lig_pattern", "Nc1ccccc1", 0),
            ("HBAcceptor.lig_pattern", "o1cccc1", 1),
            ("HBAcceptor.lig_pattern", "COC=O", 1),
            ("HBAcceptor.lig_pattern", "c1ccccc1Oc1ccccc1", 0),
            ("HBAcceptor.lig_pattern", "FC", 1),
            ("HBAcceptor.lig_pattern", "Fc1ccccc1", 1),
            ("HBAcceptor.lig_pattern", "FCF", 0),
            ("XBAcceptor.prot_pattern", "CCl", 1),
            ("XBAcceptor.prot_pattern", "c1ccccc1Cl", 1),
            ("XBAcceptor.prot_pattern", "NCl", 1),
            ("XBAcceptor.prot_pattern", "c1cccc[n+]1Cl", 1),
            ("XBAcceptor.lig_pattern", "[NH3]", 3),
            ("XBAcceptor.lig_pattern", "[NH+]C", 0),
            ("XBAcceptor.lig_pattern", "c1ccccc1", 12),
            ("Cationic.lig_pattern", "[NH4+]", 1),
            ("Cationic.lig_pattern", "[Ca+2]", 1),
            ("Cationic.lig_pattern", "CC(=[NH2+])N", 2),
            ("Cationic.lig_pattern", "NC(=[NH2+])N", 3),
            ("Cationic.prot_pattern", "[Cl-]", 1),
            ("Cationic.prot_pattern", "CC(=O)[O-]", 2),
            ("Cationic.prot_pattern", "CS(=O)[O-]", 2),
            ("Cationic.prot_pattern", "CP(=O)[O-]", 2),
            ("CationPi.cation", "[NH4+]", 1),
            ("CationPi.cation", "[Ca+2]", 1),
            ("CationPi.cation", "CC(=[NH2+])N", 2),
            ("CationPi.cation", "NC(=[NH2+])N", 3),
            ("CationPi.pi_ring", "c1ccccc1", 1),
            ("CationPi.pi_ring", "c1cocc1", 1),
            ("EdgeToFace.pi_ring", "c1ccccc1", 1),
            ("EdgeToFace.pi_ring", "c1cocc1", 1),
            ("FaceToFace.pi_ring", "c1ccccc1", 1),
            ("FaceToFace.pi_ring", "c1cocc1", 1),
            ("MetalDonor.lig_pattern", "[Mg]", 1),
            ("MetalDonor.prot_pattern", "O", 1),
            ("MetalDonor.prot_pattern", "N", 1),
            ("MetalDonor.prot_pattern", "[NH+]", 0),
            ("MetalDonor.prot_pattern", "N-C=[SH2]", 0),
            ("MetalDonor.prot_pattern", "[nH+]1ccccc1", 0),
            ("MetalDonor.prot_pattern", "Nc1ccccc1", 0),
            ("MetalDonor.prot_pattern", "o1cccc1", 0),
            ("MetalDonor.prot_pattern", "COC=O", 2),
        ],
        indirect=["interaction_qmol"],
    )
    def test_smarts_matches(self, interaction_qmol, smiles, expected):
        mol = Chem.MolFromSmiles(smiles)
        mol = Chem.AddHs(mol)
        if isinstance(interaction_qmol, list):
            n_matches = sum(
                len(mol.GetSubstructMatches(qmol)) for qmol in interaction_qmol
            )
        else:
            n_matches = len(mol.GetSubstructMatches(interaction_qmol))
        assert n_matches == expected

    @pytest.mark.parametrize(
        ["xyz", "rotation", "pi_type", "expected"],
        [
            ([0, 2.5, 4.0], [0, 0, 0], "facetoface", True),
            ([0, 3, 4.5], [0, 0, 0], "facetoface", False),
            ([0, 2, 4.5], [30, 0, 0], "facetoface", True),
            ([0, 2, 4.5], [150, 0, 0], "facetoface", True),
            ([0, 2, -4.5], [30, 0, 0], "facetoface", True),
            ([0, 2, -4.5], [150, 0, 0], "facetoface", True),
            ([1, 1.5, 3.5], [30, 15, 80], "facetoface", True),
            ([1, 2.5, 4.5], [30, 15, 65], "facetoface", True),
            ([0, 1.5, 4.5], [60, 0, 0], "edgetoface", True),
            ([0, 2, 5], [60, 0, 0], "edgetoface", True),
            ([0, 1.5, 4.5], [90, 0, 0], "edgetoface", True),
            ([0, 1.5, -4.5], [90, 0, 0], "edgetoface", True),
            ([0, 6, -0.5], [110, 0, 0], "edgetoface", True),
            ([0, 4.5, -0.5], [105, 0, 0], "edgetoface", True),
            ([0, 1.5, 4.5], [105, 0, 0], "edgetoface", False),
            ([0, 1.5, -4.5], [75, 0, 0], "edgetoface", False),
        ],
    )
    def test_pi_stacking(self, benzene, xyz, rotation, pi_type, expected, fingerprint):
        r1, r2 = self.create_rings(benzene, xyz, rotation)
        evaluate = lambda pistacking_type, r1, r2: next(
            getattr(fingerprint, pistacking_type)(r1, r2), False
        )
        assert evaluate(pi_type, r1, r2) is expected
        if expected is True:
            other = "edgetoface" if pi_type == "facetoface" else "facetoface"
            assert evaluate(other, r1, r2) is not expected
            assert evaluate("pistacking", r1, r2) is expected

    @staticmethod
    def create_rings(benzene, xyz, rotation):
        r2 = benzene.copy()
        r2.segments.segids = np.array(["U2"], dtype=object)
        tr = translate(xyz)
        rotx = rotateby(rotation[0], [1, 0, 0], ag=r2.atoms)
        roty = rotateby(rotation[1], [0, 1, 0], ag=r2.atoms)
        rotz = rotateby(rotation[2], [0, 0, 1], ag=r2.atoms)
        r2.trajectory.add_transformations(tr, rotx, roty, rotz)
        return prolif.Molecule.from_mda(benzene)[0], prolif.Molecule.from_mda(r2)[0]

    def test_edgetoface_phe331(self, ligand_mol, protein_mol):
        fp = Fingerprint()
        lig, phe331 = ligand_mol[0], protein_mol["PHE331.B"]
        assert next(fp.edgetoface(lig, phe331)) is True
        assert next(fp.pistacking(lig, phe331)) is True



"""
Reading proteins and ligands --- :mod:`prolif.molecule`
=======================================================
"""
import copy
from collections import defaultdict
from collections.abc import Sequence
from operator import attrgetter

import MDAnalysis as mda
from rdkit import Chem
from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate

from prolif.rdkitmol import BaseRDKitMol
from prolif.residue import Residue, ResidueGroup
from prolif.utils import catch_rdkit_logs, catch_warning, split_mol_by_residues


class Molecule(BaseRDKitMol):
    """Main molecule class that behaves like an RDKit :class:`~rdkit.Chem.rdchem.Mol`
    with extra attributes (see examples below). The main purpose of this class
    is to access residues as fragments of the molecule.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        A ligand or protein with a single conformer

    Attributes
    ----------
    residues : prolif.residue.ResidueGroup
        A dictionnary storing one/many :class:`~prolif.residue.Residue` indexed
        by :class:`~prolif.residue.ResidueId`. The residue list is sorted.
    n_residues : int
        Number of residues

    Examples
    --------

    .. ipython:: python
        :okwarning:

        import MDAnalysis as mda
        import prolif
        u = mda.Universe(prolif.datafiles.TOP, prolif.datafiles.TRAJ)
        mol = u.select_atoms("protein").convert_to("RDKIT")
        mol = prolif.Molecule(mol)
        mol

    You can also create a Molecule directly from a
    :class:`~MDAnalysis.core.universe.Universe`:

    .. ipython:: python
        :okwarning:

        mol = prolif.Molecule.from_mda(u, "protein")
        mol


    Notes
    -----
    Residues can be accessed easily in different ways:

    .. ipython:: python

        mol["TYR38.A"] # by resid string (residue name + number + chain)
        mol[42] # by index (from 0 to n_residues-1)
        mol[prolif.ResidueId("TYR", 38, "A")] # by ResidueId

    See :mod:`prolif.residue` for more information on residues
    """

    def __init__(self, mol):
        super().__init__(mol)
        # set mapping of atoms
        for atom in self.GetAtoms():
            atom.SetUnsignedProp("mapindex", atom.GetIdx())
        # split in residues
        residues = split_mol_by_residues(self)
        residues = [Residue(mol) for mol in residues]
        residues.sort(key=attrgetter("resid"))
        self.residues = ResidueGroup(residues)

    @classmethod
    def from_mda(cls, obj, selection=None, **kwargs):
        """Creates a Molecule from an MDAnalysis object

        Parameters
        ----------
        obj : MDAnalysis.core.universe.Universe or MDAnalysis.core.groups.AtomGroup
            The MDAnalysis object to convert
        selection : None or str
            Apply a selection to `obj` to create an AtomGroup. Uses all atoms
            in `obj` if ``selection=None``
        **kwargs : object
            Other arguments passed to the :class:`~MDAnalysis.converters.RDKit.RDKitConverter`
            of MDAnalysis

        Example
        -------
        .. ipython:: python
            :okwarning:

            mol = prolif.Molecule.from_mda(u, "protein")
            mol

        Which is equivalent to:

        .. ipython:: python
            :okwarning:

            protein = u.select_atoms("protein")
            mol = prolif.Molecule.from_mda(protein)
            mol

        """
        ag = obj.select_atoms(selection) if selection else obj.atoms
        if ag.n_atoms == 0:
            raise mda.SelectionError(f"AtomGroup is empty, please check your selection")
        mol = ag.convert_to.rdkit(**kwargs)
        return cls(mol)

    @classmethod
    def from_rdkit(cls, mol, resname="UNL", resnumber=1, chain=""):
        """Creates a Molecule from an RDKit molecule

        While directly instantiating a molecule with ``prolif.Molecule(mol)``
        would also work, this method insures that every atom is linked to an
        AtomPDBResidueInfo which is required by ProLIF

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            The input RDKit molecule
        resname : str
            The default residue name that is used if none was found
        resnumber : int
            The default residue number that is used if none was found
        chain : str
            The default chain Id that is used if none was found

        Notes
        -----
        This method only checks for an existing AtomPDBResidueInfo in the first
        atom. If none was found, it will patch all atoms with the one created
        from the method's arguments (resname, resnumber, chain).
        """
        if mol.GetAtomWithIdx(0).GetMonomerInfo():
            return cls(mol)
        mol = copy.deepcopy(mol)
        for atom in mol.GetAtoms():
            mi = Chem.AtomPDBResidueInfo(
                f" {atom.GetSymbol():<3.3}",
                residueName=resname,
                residueNumber=resnumber,
                chainId=chain,
            )
            atom.SetMonomerInfo(mi)
        return cls(mol)

    def __iter__(self):
        for residue in self.residues.values():
            yield residue

    def __getitem__(self, key):
        return self.residues[key]

    def __repr__(self):  # pragma: no cover
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        params = f"{self.n_residues} residues and {self.GetNumAtoms()} atoms"
        return f"<{name} with {params} at {id(self):#x}>"

    @property
    def n_residues(self):
        return len(self.residues)


class pdbqt_supplier(Sequence):
    """Supplies molecules, given paths to PDBQT files

    Parameters
    ----------
    paths : list
        A list (or any iterable) of PDBQT files
    template : rdkit.Chem.rdchem.Mol
        A template molecule with the correct bond orders and charges. It must
        match exactly the molecule inside the PDBQT file.
    converter_kwargs : dict
        Keyword arguments passed to the RDKitConverter of MDAnalysis
    resname : str
        Residue name for every ligand
    resnumber : int
        Residue number for every ligand
    chain : str
        Chain ID for every ligand


    Returns
    -------
    suppl : Sequence
        A sequence that provides :class:`Molecule` objects

    Example
    -------
    The supplier is typically used like this::

        >>> import glob
        >>> pdbqts = glob.glob("docking/ligand1/*.pdbqt")
        >>> lig_suppl = pdbqt_supplier(pdbqts, template)
        >>> for lig in lig_suppl:
        ...     # do something with each ligand

    .. versionchanged:: 1.0.0
        Molecule suppliers are now sequences that can be reused, indexed,
        and can return their length, instead of single-use generators.

    .. versionchanged:: 1.1.0
        Because the PDBQT supplier needs to strip hydrogen atoms before
        assigning bond orders from the template, it used to replace them
        entirely with hydrogens containing new coordinates. It now directly
        uses the hydrogen atoms present in the file and won't add explicit
        ones anymore, to prevent the fingerprint from detecting hydrogen bonds
        with "random" hydrogen atoms.
        A lot of irrelevant warnings and logs have been disabled as well.

    """

    def __init__(self, paths, template, converter_kwargs=None, **kwargs):
        self.paths = list(paths)
        self.template = template
        converter_kwargs = converter_kwargs or {}
        converter_kwargs.pop("NoImplicit", None)
        self.converter_kwargs = converter_kwargs
        self._kwargs = kwargs

    def __iter__(self):
        for pdbqt_path in self.paths:
            yield self.pdbqt_to_mol(pdbqt_path)

    def __getitem__(self, index):
        pdbqt_path = self.paths[index]
        return self.pdbqt_to_mol(pdbqt_path)

    def pdbqt_to_mol(self, pdbqt_path):
        with catch_warning(message=r"^Failed to guess the mass"):
            pdbqt = mda.Universe(pdbqt_path)
        # set attributes needed by the converter
        elements = [
            mda.topology.guessers.guess_atom_element(x) for x in pdbqt.atoms.names
        ]
        pdbqt.add_TopologyAttr("elements", elements)
        pdbqt.add_TopologyAttr("chainIDs", pdbqt.atoms.segids)
        pdbqt.atoms.types = pdbqt.atoms.elements
        # convert without infering bond orders and charges
        with catch_rdkit_logs(), catch_warning(
            message=r"^(Could not sanitize molecule)|"
            r"(No `bonds` attribute in this AtomGroup)"
        ):
            pdbqt_mol = pdbqt.atoms.convert_to.rdkit(
                NoImplicit=False, **self.converter_kwargs
            )
        mol = self._adjust_hydrogens(self.template, pdbqt_mol)
        return Molecule.from_rdkit(mol, **self._kwargs)

    @staticmethod
    def _adjust_hydrogens(template, pdbqt_mol):
        # remove explicit hydrogens and assign BO from template
        pdbqt_noH = Chem.RemoveAllHs(pdbqt_mol, sanitize=False)
        with catch_rdkit_logs():
            mol = AssignBondOrdersFromTemplate(template, pdbqt_noH)
        # mapping between pdbindex of atom bearing H --> H atom(s)
        atoms_with_hydrogens = defaultdict(list)
        for atom in pdbqt_mol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                atoms_with_hydrogens[
                    atom.GetNeighbors()[0].GetIntProp("_MDAnalysis_index")
                ].append(atom)
        # mapping between atom that should be bearing a H in RWMol and corresponding H(s)
        reverse_mapping = {}
        for atom in mol.GetAtoms():
            if (idx := atom.GetIntProp("_MDAnalysis_index")) in atoms_with_hydrogens:
                reverse_mapping[atom.GetIdx()] = atoms_with_hydrogens[idx]
                atom.SetNumExplicitHs(0)
        # add missing Hs
        pdb_conf = pdbqt_mol.GetConformer()
        rwmol = Chem.RWMol(mol)
        mol_conf = rwmol.GetConformer()
        for atom_idx, hydrogens in reverse_mapping.items():
            for hydrogen in hydrogens:
                h_idx = rwmol.AddAtom(hydrogen)
                xyz = pdb_conf.GetAtomPosition(hydrogen.GetIdx())
                mol_conf.SetAtomPosition(h_idx, xyz)
                rwmol.AddBond(atom_idx, h_idx, Chem.BondType.SINGLE)
        mol = rwmol.GetMol()
        # sanitize
        mol.UpdatePropertyCache()
        Chem.SanitizeMol(mol)
        return mol

    def __len__(self):
        return len(self.paths)


class sdf_supplier(Sequence):
    """Supplies molecules, given a path to an SDFile

    Parameters
    ----------
    path : str
        A path to the .sdf file
    resname : str
        Residue name for every ligand
    resnumber : int
        Residue number for every ligand
    chain : str
        Chain ID for every ligand

    Returns
    -------
    suppl : Sequence
        A sequence that provides :class:`Molecule` objects. Can be indexed

    Example
    -------
    The supplier is typically used like this::

        >>> lig_suppl = sdf_supplier("docking/output.sdf")
        >>> for lig in lig_suppl:
        ...     # do something with each ligand

    .. versionchanged:: 1.0.0
        Molecule suppliers are now sequences that can be reused, indexed,
        and can return their length, instead of single-use generators.

    """

    def __init__(self, path, **kwargs):
        self.path = path
        self._suppl = Chem.SDMolSupplier(path, removeHs=False)
        self._kwargs = kwargs

    def __iter__(self):
        for mol in self._suppl:
            yield Molecule.from_rdkit(mol, **self._kwargs)

    def __getitem__(self, index):
        mol = self._suppl[index]
        return Molecule.from_rdkit(mol, **self._kwargs)

    def __len__(self):
        return len(self._suppl)


class mol2_supplier(Sequence):
    """Supplies molecules, given a path to a MOL2 file

    Parameters
    ----------
    path : str
        A path to the .mol2 file
    resname : str
        Residue name for every ligand
    resnumber : int
        Residue number for every ligand
    chain : str
        Chain ID for every ligand

    Returns
    -------
    suppl : Sequence
        A sequence that provides :class:`Molecule` objects

    Example
    -------
    The supplier is typically used like this::

        >>> lig_suppl = mol2_supplier("docking/output.mol2")
        >>> for lig in lig_suppl:
        ...     # do something with each ligand

    .. versionchanged:: 1.0.0
        Molecule suppliers are now sequences that can be reused, indexed,
        and can return their length, instead of single-use generators.

    """

    def __init__(self, path, **kwargs):
        self.path = path
        self._kwargs = kwargs

    def __iter__(self):
        block = []
        with open(self.path, "r") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                if block and line.startswith("@<TRIPOS>MOLECULE"):
                    yield self.block_to_mol(block)
                    block = []
                block.append(line)
            yield self.block_to_mol(block)

    def block_to_mol(self, block):
        mol = Chem.MolFromMol2Block("".join(block), removeHs=False)
        return Molecule.from_rdkit(mol, **self._kwargs)

    def __getitem__(self, index):
        if index < 0:
            index %= len(self)
        mol_index = -1
        molblock_started = False
        block = []
        with open(self.path, "r") as f:
            for line in f:
                if line.startswith("@<TRIPOS>MOLECULE"):
                    mol_index += 1
                    if mol_index > index:
                        return self.block_to_mol(block)
                    if mol_index == index:
                        molblock_started = True
                if molblock_started:
                    block.append(line)
        if block:
            return self.block_to_mol(block)
        raise ValueError(f"Could not parse molecule with index {index}")

    def __len__(self):
        with open(self.path, "r") as f:
            n_mols = sum(line.startswith("@<TRIPOS>MOLECULE") for line in f)
        return n_mols



from prolif import datafiles
from prolif._version import __version__
from prolif.fingerprint import Fingerprint
from prolif.molecule import Molecule, mol2_supplier, pdbqt_supplier, sdf_supplier
from prolif.plotting.residues import display_residues
from prolif.residue import ResidueId
from prolif.utils import get_residues_near_ligand, to_bitvectors, to_dataframe



"""
Storing interactions --- :mod:`prolif.ifp`
==========================================
"""
from collections import UserDict

from prolif.residue import ResidueId


class IFP(UserDict):
    """Mapping between residue pairs and interaction fingerprint.

    Notes
    -----
    This class provides an easy way to access interaction data from the
    :attr:`~prolif.fingerprint.Fingerprint.ifp` dictionary. This class is a dictionary
    formatted as:

    .. code-block:: text

        {
            tuple[<residue_id>, <residue_id>]: {
                <interaction name>: tuple[{
                    "indices": {
                        "ligand": tuple[int, ...],
                        "protein": tuple[int, ...]
                    },
                    "parent_indices": {
                        "ligand": tuple[int, ...],
                        "protein": tuple[int, ...]
                    },
                    <other metadata>: <value>
                }, ...]
            }
        }

    Here ``<residue_id>`` corresponds to a :class:`~prolif.residue.ResidueId` object.
    For convenience, one can directly use strings rather than ``ResidueId`` objects when
    indexing the IFP, e.g. ``ifp[("LIG1.G", "ASP129.A")]``.
    You can also use a single ``ResidueId`` or string to return a filtered IFP that only
    contains interactions with the specified residue, e.g. ``ifp["ASP129.A"]``.
    """

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError as exc:
            if isinstance(key, tuple) and len(key) == 2:
                lig_res, prot_res = key
                if isinstance(lig_res, ResidueId):
                    raise exc from None
                key = ResidueId.from_string(lig_res), ResidueId.from_string(prot_res)
                return self.data[key]
            if isinstance(key, str):
                key = ResidueId.from_string(key)
            if isinstance(key, ResidueId):
                return IFP(
                    {
                        residue_tuple: interactions
                        for residue_tuple, interactions in self.data.items()
                        if key in residue_tuple
                    }
                )
        raise KeyError(
            f"{key} does not correspond to a valid IFP key: it must be a tuple of "
            "either ResidueId or residue string. If you need to filter the IFP, a "
            "single ResidueId or residue string can also be used."
        )



import pytest
from numpy.testing import assert_equal
from rdkit import Chem
from rdkit.Chem import AllChem

from prolif.residue import Residue, ResidueGroup, ResidueId

from .test_base import TestBaseRDKitMol


class TestResidueId:
    @pytest.mark.parametrize(
        "name, number, chain",
        [
            ("ALA", None, None),
            ("ALA", 1, None),
            ("ALA", 0, None),
            ("ALA", None, "B"),
            ("ALA", 1, "B"),
            (None, 1, "B"),
            (None, None, "B"),
            (None, 1, None),
            (None, None, None),
            ("", None, None),
            (None, None, ""),
            ("", None, ""),
        ],
    )
    def test_init(self, name, number, chain):
        resid = ResidueId(name, number, chain)
        name = name or "UNK"
        number = number or 0
        chain = chain or None
        assert resid.name == name
        assert resid.number == number
        assert resid.chain == chain

    @pytest.mark.parametrize(
        "name, number, chain",
        [
            ("ALA", None, None),
            ("ALA", 1, None),
            ("ALA", 0, None),
            ("ALA", None, "B"),
            ("ALA", 1, "B"),
            ("DA", 1, None),
            (None, 1, "B"),
            (None, None, "B"),
            (None, 1, None),
            (None, None, None),
            ("", None, None),
            (None, None, ""),
            ("", None, ""),
        ],
    )
    def test_from_atom(self, name, number, chain):
        atom = Chem.Atom(1)
        mi = Chem.AtomPDBResidueInfo()
        if name is not None:
            mi.SetResidueName(name)
        if number is not None:
            mi.SetResidueNumber(number)
        if chain is not None:
            mi.SetChainId(chain)
        atom.SetMonomerInfo(mi)
        resid = ResidueId.from_atom(atom)
        name = name or "UNK"
        number = number or 0
        chain = chain or None
        assert resid.name == name
        assert resid.number == number
        assert resid.chain == chain

    def test_from_atom_no_mi(self):
        atom = Chem.Atom(1)
        resid = ResidueId.from_atom(atom)
        assert resid.name == "UNK"
        assert resid.number == 0
        assert resid.chain is None

    @pytest.mark.parametrize(
        "resid_str, expected",
        [
            ("ALA", ("ALA", 0, None)),
            ("ALA1", ("ALA", 1, None)),
            ("ALA.B", ("ALA", 0, "B")),
            ("ALA1.B", ("ALA", 1, "B")),
            ("1.B", ("UNK", 1, "B")),
            (".B", ("UNK", 0, "B")),
            (".0", ("UNK", 0, "0")),
            ("1", ("UNK", 1, None)),
            ("", ("UNK", 0, None)),
            ("DA2.A", ("DA", 2, "A")),
            ("DA2", ("DA", 2, None)),
            ("DA", ("DA", 0, None)),
        ],
    )
    def test_from_string(self, resid_str, expected):
        resid = ResidueId.from_string(resid_str)
        assert resid == ResidueId(*expected)

    def test_eq(self):
        name, number, chain = "ALA", 1, "A"
        res1 = ResidueId(name, number, chain)
        res2 = ResidueId(name, number, chain)
        assert res1 == res2

    @pytest.mark.parametrize(
        "res1, res2",
        [
            ("ALA1.A", "ALA1.B"),
            ("ALA2.A", "ALA3.A"),
            ("ALA4.A", "ALA1.B"),
        ],
    )
    def test_lt(self, res1, res2):
        res1 = ResidueId.from_string(res1)
        res2 = ResidueId.from_string(res2)
        assert res1 < res2

    @pytest.mark.parametrize("resid_str", ["ALA1.A", "DA2.B", "HIS3", "UNK0"])
    def test_repr(self, resid_str):
        resid = ResidueId.from_string(resid_str)
        expected = f"ResidueId({resid.name}, {resid.number}, {resid.chain})"
        assert repr(resid) == expected
        assert str(resid) == resid_str


class TestResidue(TestBaseRDKitMol):
    @pytest.fixture(scope="class")
    def mol(self):
        mol = Chem.MolFromSequence("A")
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol)
        return Residue(mol)

    def test_init(self, mol):
        resid = ResidueId.from_atom(mol.GetAtomWithIdx(0))
        assert mol.resid == resid

    def test_str(self, mol):
        assert str(mol) == "ALA1.A"


class TestResidueGroup:
    @pytest.fixture(scope="class")
    def residues(self):
        sequence = "ARNDCQEGHILKMFPSTWYV"
        protein = Chem.MolFromSequence(sequence)
        residues = [
            Residue(res) for res in Chem.SplitMolByPDBResidues(protein).values()
        ]
        return residues

    def test_init(self, residues):
        rg = ResidueGroup(residues)
        for rg_res, res in zip(rg._residues, residues):
            assert rg_res is res
        for (resid, rg_res), res in zip(rg.items(), residues):
            assert rg_res is res
            assert resid is rg_res.resid
        resinfo = [(r.resid.name, r.resid.number, r.resid.chain) for r in residues]
        name, number, chain = zip(*resinfo)
        assert_equal(rg.name, name)
        assert_equal(rg.number, number)
        assert_equal(rg.chain, chain)

    def test_init_empty(self):
        rg = ResidueGroup([])
        assert_equal(rg.name, [])
        assert_equal(rg.number, [])
        assert_equal(rg.chain, [])
        assert_equal(rg._residues, [])
        assert rg.data == {}

    def test_n_residues(self, residues):
        rg = ResidueGroup(residues)
        assert rg.n_residues == len(rg)
        assert rg.n_residues == 20

    @pytest.mark.parametrize(
        "ix, resid, resid_str",
        [
            (0, ("ALA", 1, "A"), "ALA1.A"),
            (4, ("CYS", 5, "A"), "CYS5.A"),
            (6, ("GLU", 7, "A"), "GLU7.A"),
            (9, ("ILE", 10, "A"), "ILE10.A"),
            (19, ("VAL", 20, "A"), "VAL20.A"),
            (-1, ("VAL", 20, "A"), "VAL20.A"),
        ],
    )
    def test_getitem(self, residues, ix, resid, resid_str):
        rg = ResidueGroup(residues)
        resid = ResidueId(*resid)
        assert rg[ix] == rg[resid]
        assert rg[ix] == rg[resid_str]

    def test_getitem_keyerror(self):
        rg = ResidueGroup([])
        with pytest.raises(KeyError, match="Expected a ResidueId, int, or str"):
            rg[True]
        with pytest.raises(KeyError, match="Expected a ResidueId, int, or str"):
            rg[1.5]

    def test_select(self, protein_mol):
        rg = protein_mol.residues
        assert rg.select(rg.name == "TYR").n_residues == 6
        assert rg.select(rg.number == 109).n_residues == 1
        assert rg.select(rg.number == 1).n_residues == 0
        assert rg.select(rg.chain == "A").n_residues == 36
        # and
        assert rg.select((rg.chain == "A") & (rg.name == "TYR")).n_residues == 5
        # or
        assert rg.select((rg.chain == "A") | (rg.name == "TYR")).n_residues == 37
        # xor
        assert rg.select((rg.chain == "A") ^ (rg.name == "TYR")).n_residues == 32
        # not
        assert rg.select(~(rg.chain == "A")).n_residues == 20

    def test_select_sameas_getitem(self, protein_mol):
        rg = protein_mol.residues
        sel = rg.select((rg.name == "TYR") & (rg.number == 38))[0]
        assert sel.resid is rg["TYR38.A"].resid



import pytest
from numpy.testing import assert_array_almost_equal
from rdkit import Chem
from rdkit.Chem.rdMolTransforms import ComputeCentroid

from prolif.rdkitmol import BaseRDKitMol


class TestBaseRDKitMol:
    @pytest.fixture(scope="class")
    def mol(self, rdkit_mol):
        return BaseRDKitMol(rdkit_mol)

    def test_init(self, mol):
        assert isinstance(mol, Chem.Mol)

    def test_centroid(self, mol):
        expected = ComputeCentroid(mol.GetConformer())
        assert_array_almost_equal(mol.centroid, expected)

    def test_xyz(self, mol):
        expected = mol.GetConformer().GetPositions()
        assert_array_almost_equal(mol.xyz, expected)



"""
Calculate a Protein-Ligand Interaction Fingerprint --- :mod:`prolif.fingerprint`
================================================================================

.. ipython:: python
    :okwarning:

    import MDAnalysis as mda
    from rdkit.DataStructs import TanimotoSimilarity
    import prolif
    u = mda.Universe(prolif.datafiles.TOP, prolif.datafiles.TRAJ)
    prot = u.select_atoms("protein")
    lig = u.select_atoms("resname LIG")
    fp = prolif.Fingerprint(["HBDonor", "HBAcceptor", "PiStacking", "CationPi",
                             "Cationic"])
    fp.run(u.trajectory[:5], lig, prot)
    df = fp.to_dataframe()
    df
    bv = fp.to_bitvectors()
    TanimotoSimilarity(bv[0], bv[1])
    # save results
    fp.to_pickle("fingerprint.pkl")
    # load
    fp = prolif.Fingerprint.from_pickle("fingerprint.pkl")

"""
import warnings
from collections.abc import Sized
from functools import wraps
from typing import Literal, Optional, Tuple

import dill
import multiprocess as mp
import numpy as np
from rdkit import Chem
from tqdm.auto import tqdm

from prolif.ifp import IFP
from prolif.interactions.base import _BASE_INTERACTIONS, _INTERACTIONS
from prolif.molecule import Molecule
from prolif.parallel import MolIterablePool, TrajectoryPool
from prolif.plotting.utils import IS_NOTEBOOK
from prolif.utils import (
    get_residues_near_ligand,
    to_bitvectors,
    to_countvectors,
    to_dataframe,
)


def first_occurence(interaction):
    @wraps(interaction)
    def wrapped(*args, **kwargs):
        return next(interaction(*args, **kwargs), None)

    return wrapped


def all_occurences(interaction):
    @wraps(interaction)
    def wrapped(*args, **kwargs):
        return tuple(interaction(*args, **kwargs))

    return wrapped


class Fingerprint:
    """Class that generates an interaction fingerprint between two molecules

    While in most cases the fingerprint will be generated between a ligand and
    a protein, it is also possible to use this class for protein-protein
    interactions, or host-guest systems.

    Parameters
    ----------
    interactions : list
        List of names (str) of interaction classes as found in the
        :mod:`prolif.interactions` module.
    parameters : dict, optional
        New parameters for the interactions. Mapping between an interaction name and a
        dict of parameters as they appear in the interaction class.
    count : bool
        For a given interaction class and pair of residues, there might be multiple
        combinations of atoms that satisfy the interaction constraints. This parameter
        controls how the fingerprint treats these different combinations:

        * ``False``: returns the first combination that satisfy the constraints (fast)
        * ``True``: returns all the combinations (slow)

    vicinity_cutoff : float
        Automatically restrict the analysis to residues within this range of the ligand.
        This parameter is ignored if the ``residues`` parameter of the ``run`` methods
        is set to anything other than ``None``.

    Attributes
    ----------
    interactions : dict
        Dictionary of interaction classes indexed by class name. For more
        details, see :mod:`prolif.interactions`
    n_interactions : int
        Number of interaction functions registered by the fingerprint
    vicinity_cutoff : float
        Used when calling :func:`prolif.utils.get_residues_near_ligand`.
    count : bool
        Whether to keep track of all interaction occurences or just the first one.
    ifp : dict, optional
        Dict of interaction fingerprints in a sparse format for the given trajectory or
        docking poses: ``{<frame number>: <IFP>}``. See the :class:`~prolif.ifp.IFP`
        class for more information.

    Raises
    ------
    NameError
        Unknown interaction in the ``interactions`` or ``parameters`` parameters.

    Notes
    -----
    You can use the fingerprint generator in multiple ways:

    - On a trajectory directly from MDAnalysis objects:

    .. ipython:: python

        prot = u.select_atoms("protein")
        lig = u.select_atoms("resname LIG")
        fp = prolif.Fingerprint(["HBDonor", "HBAcceptor", "PiStacking",
                                 "Hydrophobic"])
        fp.run(u.trajectory[:5], lig, prot)
        fp.to_dataframe()

    - On two single structures (from RDKit or MDAnalysis):

    .. ipython:: python

        u.trajectory[0]  # use coordinates of the first frame
        prot = prolif.Molecule.from_mda(prot)
        lig = prolif.Molecule.from_mda(lig)
        ifp = fp.generate(lig, prot, metadata=True)
        prolif.to_dataframe({0: ifp}, fp.interactions)

    - On a specific pair of residues for a specific interaction:

    .. ipython:: python

        # ligand-protein
        next(fp.hbdonor(lig, prot["ASP129.A"]))
        # protein-protein
        next(fp.hbacceptor(prot["ASP129.A"], prot["CYS133.A"]))

    You can also obtain the indices of atoms responsible for the interaction:

    .. ipython:: python

        fp.metadata(lig, prot["ASP129.A"])
        next(fp.hbdonor(lig, prot["ASP129.A"], metadata=True), None)


    .. versionchanged:: 1.0.0
        Added pickle support

    .. versionchanged:: 2.0.0
        Changed the format of the :attr:`~Fingerprint.ifp` attribute to be a dictionary
        containing more complete interaction metadata instead of just atom indices.
        Removed the ``return_atoms`` argument in :meth:`~Fingerprint.to_dataframe`.
        Users should directly use :attr:`~Fingerprint.ifp` instead.
        Added the :meth:`~Fingerprint.plot_lignetwork` method to generate the
        :class:`~prolif.plotting.network.LigNetwork` plot.
        Replaced the ``Fingerprint.bitvector_atoms`` method with
        :meth:`Fingerprint.metadata`.
        Added a ``vicinity_cutoff`` parameter controlling the distance used
        to automatically restrict the IFP calculation to residues within the specified
        range of the ligand.
        Removed the ``__wrapped__`` attribute on interaction methods that are available
        from the fingerprint object. These methods now accept a ``metadata`` parameter
        instead.
        Added the ``parameters`` argument to set the interaction classes parameters
        without having to create a new class.
        Added the ``count`` argument to control for a given pair of residues whether to
        return all occurences of an interaction or only the first one.

    """

    def __init__(
        self,
        interactions=[
            "Hydrophobic",
            "HBDonor",
            "HBAcceptor",
            "PiStacking",
            "Anionic",
            "Cationic",
            "CationPi",
            "PiCation",
            "VdWContact",
        ],
        parameters=None,
        count=False,
        vicinity_cutoff=6.0,
    ):
        self.count = count
        self._set_interactions(interactions, parameters)
        self.vicinity_cutoff = vicinity_cutoff

    def _set_interactions(self, interactions, parameters):
        # read interactions to compute
        parameters = parameters or {}
        if interactions == "all":
            interactions = self.list_available()
        # sanity check
        self._check_valid_interactions(interactions, "interactions")
        self._check_valid_interactions(parameters, "parameters")
        # add interaction methods
        self.interactions = {}
        wrapper = all_occurences if self.count else first_occurence
        for name, interaction_cls in _INTERACTIONS.items():
            # create instance with custom parameters if available
            interaction = interaction_cls(**parameters.get(name, {}))
            setattr(self, name.lower(), interaction)
            if name in interactions:
                self.interactions[name] = wrapper(interaction)

    def _check_valid_interactions(self, interactions_iterable, varname):
        """Raises a NameError if an unknown interaction is given."""
        unsafe = set(interactions_iterable)
        unknown = unsafe.symmetric_difference(_INTERACTIONS.keys()) & unsafe
        if unknown:
            raise NameError(
                f"Unknown interaction(s) in {varname!r}: {', '.join(unknown)}"
            )

    def __repr__(self):  # pragma: no cover
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        params = f"{self.n_interactions} interactions: {list(self.interactions.keys())}"
        return f"<{name}: {params} at {id(self):#x}>"

    @staticmethod
    def list_available(show_hidden=False):
        """List interactions available to the Fingerprint class.

        Parameters
        ----------
        show_hidden : bool
            Show hidden classes (base classes meant to be inherited from to create
            custom interactions).
        """
        if show_hidden:
            return sorted(_BASE_INTERACTIONS) + sorted(_INTERACTIONS)
        return sorted(_INTERACTIONS)

    @property
    def n_interactions(self):
        return len(self.interactions)

    def bitvector(self, res1, res2):
        """Generates the complete bitvector for the interactions between two
        residues. To access metadata for each interaction, see
        :meth:`~Fingerprint.metadata`.

        Parameters
        ----------
        res1 : prolif.residue.Residue
            A residue, usually from a ligand
        res2 : prolif.residue.Residue
            A residue, usually from a protein

        Returns
        -------
        bitvector : numpy.ndarray
            An array storing the encoded interactions between res1 and res2. Depending
            on :attr:`Fingerprint.count`, the dtype of the array will be bool or uint8.
        """
        bitvector = [
            interaction(res1, res2) for interaction in self.interactions.values()
        ]
        if self.count:
            return np.array([sum(bits) for bits in bitvector], dtype=np.uint8)
        return np.array(bitvector, dtype=bool)

    def metadata(self, res1, res2):
        """Generates a metadata dictionary for the interactions between two residues.

        Parameters
        ----------
        res1 : prolif.residue.Residue
            A residue, usually from a ligand
        res2 : prolif.residue.Residue
            A residue, usually from a protein

        Returns
        -------
        metadata : dict[str, tuple[dict, ...]]
            Dict containing tuples of metadata dictionaries indexed by interaction
            name. If a specific interaction is not present between residues, it is
            filtered out of the dictionary.


        .. versionchanged:: 0.3.2
            Atom indices are returned as two separate lists instead of a single
            list of tuples

        .. versionchanged:: 2.0.0
            Returns a dictionnary with all available metadata for each interaction,
            rather than just atom indices.

        """
        return {
            name: metadata if self.count else (metadata,)
            for name, interaction in self.interactions.items()
            if (metadata := interaction(res1, res2, metadata=True))
        }

    def generate(self, lig, prot, residues=None, metadata=False):
        """Generates the interaction fingerprint between 2 molecules

        Parameters
        ----------
        lig : prolif.molecule.Molecule
            Molecule for the ligand
        prot : prolif.molecule.Molecule
            Molecule for the protein
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue (see :attr:`~Fingerprint.vicinity_cutoff`)
        metadata : bool
            For each residue pair and interaction, return an interaction metadata
            dictionary instead of bits.

        Returns
        -------
        ifp : prolif.ifp.IFP
            A dictionary indexed by ``(ligand, protein)`` residue pairs. The
            format for values will depend on ``metadata``:

            - A single numpy array if ``metadata=False``
            - A sparse dictionary of metadata tuples indexed by interaction name if
              ``metadata=True``

        Example
        -------
        ::

            >>> u = mda.Universe("complex.pdb")
            >>> lig = prolif.Molecule.from_mda(u, "resname LIG")
            >>> prot = prolif.Molecule.from_mda(u, "protein")
            >>> fp = prolif.Fingerprint()
            >>> ifp = fp.generate(lig, prot)

        .. versionadded:: 0.3.2

        .. versionchanged:: 2.0.0
            ``return_atoms`` replaced by ``metadata``, and it now returns a sparse
            dictionary of metadata tuples indexed by interaction name instead of a
            tuple of arrays.
        """
        ifp = IFP()
        prot_residues = prot.residues if residues == "all" else residues
        get_interactions = self.metadata if metadata else self.bitvector
        for lresid, lres in lig.residues.items():
            if residues is None:
                prot_residues = get_residues_near_ligand(
                    lres, prot, self.vicinity_cutoff
                )
            for prot_key in prot_residues:
                pres = prot[prot_key]
                key = (lresid, pres.resid)
                interactions = get_interactions(lres, pres)
                if any(interactions):
                    ifp[key] = interactions
        return ifp

    def run(
        self,
        traj,
        lig,
        prot,
        *,
        residues=None,
        converter_kwargs=None,
        progress=True,
        n_jobs=None,
    ):
        """Generates the fingerprint on a trajectory for a ligand and a protein

        Parameters
        ----------
        traj : MDAnalysis.coordinates.base.ProtoReader or MDAnalysis.coordinates.base.FrameIteratorSliced
            Iterate over this Universe trajectory or sliced trajectory object
            to extract the frames used for the fingerprint extraction
        lig : MDAnalysis.core.groups.AtomGroup
            An MDAnalysis AtomGroup for the ligand
        prot : MDAnalysis.core.groups.AtomGroup
            An MDAnalysis AtomGroup for the protein (with multiple residues)
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue.
        converter_kwargs : tuple[dict, dict], optional
            Tuple of kwargs passed to the underlying :class:`~MDAnalysis.converters.RDKit.RDKitConverter`
            from MDAnalysis: the first for the ligand, and the second for the protein
        progress : bool
            Display a :class:`~tqdm.std.tqdm` progressbar while running the calculation
        n_jobs : int or None
            Number of processes to run in parallel. If ``n_jobs=None``, the
            analysis will use all available CPU threads, while if ``n_jobs=1``,
            the analysis will run in serial.

        Raises
        ------
        ValueError
            If ``n_jobs <= 0``

        Returns
        -------
        prolif.fingerprint.Fingerprint
            The Fingerprint instance that generated the fingerprint

        Example
        -------
        ::

            >>> u = mda.Universe("top.pdb", "traj.nc")
            >>> lig = u.select_atoms("resname LIG")
            >>> prot = u.select_atoms("protein")
            >>> fp = prolif.Fingerprint().run(u.trajectory[:10], lig, prot)

        .. seealso::

            - :meth:`Fingerprint.generate` to generate the fingerprint between
              two single structures.
            - :meth:`Fingerprint.run_from_iterable` to generate the fingerprint
              between a protein and a collection of ligands.

        .. versionchanged:: 0.3.2
            Moved the ``return_atoms`` parameter from the ``run`` method to the
            dataframe conversion code

        .. versionchanged:: 1.0.0
            Added support for multiprocessing

        .. versionchanged:: 1.1.0
            Added support for passing kwargs to the RDKitConverter through
            the ``converter_kwargs`` parameter

        .. versionchanged:: 2.0.0
            Changed the format of the :attr:`~Fingerprint.ifp` attribute to be a
            dictionary containing more complete interaction metadata instead of just
            atom indices.

        """
        if n_jobs is not None and n_jobs < 1:
            raise ValueError("n_jobs must be > 0 or None")
        if converter_kwargs is not None and len(converter_kwargs) != 2:
            raise ValueError("converter_kwargs must be a list of 2 dicts")

        converter_kwargs = converter_kwargs or ({}, {})
        if residues == "all":
            residues = list(Molecule.from_mda(prot, **converter_kwargs[1]).residues)
        if n_jobs != 1:
            return self._run_parallel(
                traj,
                lig,
                prot,
                residues=residues,
                converter_kwargs=converter_kwargs,
                progress=progress,
                n_jobs=n_jobs,
            )

        iterator = tqdm(traj) if progress else traj
        ifp = {}
        for ts in iterator:
            lig_mol = Molecule.from_mda(lig, **converter_kwargs[0])
            prot_mol = Molecule.from_mda(prot, **converter_kwargs[1])
            ifp[int(ts.frame)] = self.generate(
                lig_mol, prot_mol, residues=residues, metadata=True
            )
        self.ifp = ifp
        return self

    def _run_parallel(
        self,
        traj,
        lig,
        prot,
        residues=None,
        converter_kwargs=None,
        progress=True,
        n_jobs=None,
    ):
        """Parallel implementation of :meth:`~Fingerprint.run`"""
        n_chunks = n_jobs if n_jobs else mp.cpu_count()
        try:
            n_frames = traj.n_frames
        except AttributeError:
            # sliced trajectory
            frames = range(traj.start, traj.stop, traj.step)
            traj = lig.universe.trajectory
        else:
            frames = range(n_frames)
        chunks = np.array_split(frames, n_chunks)
        args_iterable = [(traj, lig, prot, chunk) for chunk in chunks]
        ifp = {}

        with TrajectoryPool(
            n_jobs,
            fingerprint=self,
            residues=residues,
            tqdm_kwargs={"total": len(frames), "disable": not progress},
            rdkitconverter_kwargs=converter_kwargs,
        ) as pool:
            for ifp_data_chunk in pool.process(args_iterable):
                ifp.update(ifp_data_chunk)

        self.ifp = ifp
        return self

    def run_from_iterable(
        self, lig_iterable, prot_mol, *, residues=None, progress=True, n_jobs=None
    ):
        """Generates the fingerprint between a list of ligands and a protein

        Parameters
        ----------
        lig_iterable : list or generator
            An iterable yielding ligands as :class:`~prolif.molecule.Molecule`
            objects
        prot_mol : prolif.molecule.Molecule
            The protein
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue.
        progress : bool
            Display a :class:`~tqdm.std.tqdm` progressbar while running the calculation
        n_jobs : int or None
            Number of processes to run in parallel. If ``n_jobs=None``, the
            analysis will use all available CPU threads, while if ``n_jobs=1``,
            the analysis will run in serial.

        Raises
        ------
        ValueError
            If ``n_jobs <= 0``

        Returns
        -------
        prolif.fingerprint.Fingerprint
            The Fingerprint instance that generated the fingerprint

        Example
        -------
        ::

            >>> prot = mda.Universe("protein.pdb")
            >>> prot = prolif.Molecule.from_mda(prot)
            >>> lig_iter = prolif.mol2_supplier("docking_output.mol2")
            >>> fp = prolif.Fingerprint()
            >>> fp.run_from_iterable(lig_iter, prot)

        .. seealso::

            :meth:`Fingerprint.generate` to generate the fingerprint between
            two single structures

        .. versionchanged:: 0.3.2
            Moved the ``return_atoms`` parameter from the ``run_from_iterable``
            method to the dataframe conversion code

        .. versionchanged:: 1.0.0
            Added support for multiprocessing

        .. versionchanged:: 2.0.0
            Changed the format of the :attr:`~Fingerprint.ifp` attribute to be a
            dictionary containing more complete interaction metadata instead of just
            atom indices.

        """
        if n_jobs is not None and n_jobs < 1:
            raise ValueError("n_jobs must be > 0 or None")
        if residues == "all":
            residues = list(prot_mol.residues)
        if n_jobs != 1:
            return self._run_iter_parallel(
                lig_iterable=lig_iterable,
                prot_mol=prot_mol,
                residues=residues,
                progress=progress,
                n_jobs=n_jobs,
            )

        iterator = tqdm(lig_iterable) if progress else lig_iterable
        ifp = {}
        for i, lig_mol in enumerate(iterator):
            ifp[i] = self.generate(lig_mol, prot_mol, residues=residues, metadata=True)
        self.ifp = ifp
        return self

    def _run_iter_parallel(
        self, lig_iterable, prot_mol, residues=None, progress=True, n_jobs=None
    ):
        """Parallel implementation of :meth:`~Fingerprint.run_from_iterable`"""
        total = (
            len(lig_iterable)
            if isinstance(lig_iterable, (Chem.SDMolSupplier, Sized))
            else None
        )
        ifp = {}

        with MolIterablePool(
            n_jobs,
            fingerprint=self,
            prot_mol=prot_mol,
            residues=residues,
            tqdm_kwargs={"total": total, "disable": not progress},
        ) as pool:
            for i, ifp_data in enumerate(pool.process(lig_iterable)):
                ifp[i] = ifp_data

        self.ifp = ifp
        return self

    def to_dataframe(
        self, *, count=None, dtype=None, drop_empty=True, index_col="Frame"
    ):
        """Converts fingerprints to a pandas DataFrame

        Parameters
        ----------
        count : bool or None
            Whether to output a count fingerprint or not.
        dtype : object or None
            Cast the dataframe values to this type. If ``None``, uses ``np.uint8`` if
            ``count=True``, else ``bool``.
        drop_empty : bool
            Drop columns with only empty values
        index_col : str
            Name of the index column in the DataFrame

        Returns
        -------
        df : pandas.DataFrame
            A DataFrame storing the results frame by frame and residue by
            residue. See :meth:`~prolif.utils.to_dataframe` for more
            information

        Raises
        ------
        AttributeError
            If the :meth:`run` method hasn't been used

        Example
        -------
        ::

            >>> df = fp.to_dataframe(dtype=np.uint8)
            >>> print(df)
            ligand             LIG1.G
            protein             ILE59                  ILE55       TYR93
            interaction   Hydrophobic HBAcceptor Hydrophobic Hydrophobic PiStacking
            Frame
            0                       0          1           0           0          0
            ...

        .. versionchanged:: 2.0.0
            Removed the ``return_atoms`` parameter. You can access more metadata
            information directly through :attr:`~Fingerprint.ifp`. Added the ``count``
            parameter.
        """
        if hasattr(self, "ifp"):
            return to_dataframe(
                self.ifp,
                self.interactions,
                count=self.count if count is None else count,
                dtype=dtype,
                drop_empty=drop_empty,
                index_col=index_col,
            )
        raise AttributeError("Please use the `run` method before")

    def to_bitvectors(self):
        """Converts fingerprints to a list of RDKit ExplicitBitVector

        Returns
        -------
        bvs : list
            A list of :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect`
            for each frame

        Raises
        ------
        AttributeError
            If the :meth:`run` method hasn't been used

        Example
        -------
        ::

            >>> from rdkit.DataStructs import TanimotoSimilarity
            >>> bv = fp.to_bitvectors()
            >>> TanimotoSimilarity(bv[0], bv[1])
            0.42

        """
        if hasattr(self, "ifp"):
            df = self.to_dataframe()
            return to_bitvectors(df)
        raise AttributeError("Please use the `run` method before")

    def to_countvectors(self):
        """Converts fingerprints to a list of RDKit UIntSparseIntVect

        Returns
        -------
        cvs : list
            A list of :class:`~rdkit.DataStructs.cDataStructs.UIntSparseIntVect`
            for each frame

        Raises
        ------
        AttributeError
            If the :meth:`run` method hasn't been used

        Example
        -------
        ::

            >>> from rdkit.DataStructs import TanimotoSimilarity
            >>> cv = fp.to_countvectors()
            >>> TanimotoSimilarity(cv[0], cv[1])
            0.42


        .. versionadded: 2.0.0
        """
        if hasattr(self, "ifp"):
            df = self.to_dataframe()
            return to_countvectors(df)
        raise AttributeError("Please use the `run` method before")

    def to_pickle(self, path=None):
        """Dumps the fingerprint object as a pickle.

        Parameters
        ----------
        path : str, pathlib.Path or None
            Output path. If ``None``, the method returns the pickle as bytes.

        Returns
        -------
        obj : None or bytes
            ``None`` if ``path`` is set, else the bytes corresponding to the
            pickle

        Example
        -------
        ::

            >>> dump = fp.to_pickle()
            >>> saved_fp = Fingerprint.from_pickle(dump)
            >>> fp.to_pickle("data/fp.pkl")
            >>> saved_fp = Fingerprint.from_pickle("data/fp.pkl")

        .. seealso::

            :meth:`~Fingerprint.from_pickle` for loading the pickle dump.

        .. versionadded:: 1.0.0

        .. versionchanged:: 2.0.0
            Switched to dill instead of pickle

        """
        if path:
            with open(path, "wb") as f:
                dill.dump(self, f)
            return None
        return dill.dumps(self)

    @classmethod
    def from_pickle(cls, path_or_bytes):
        """Creates a fingerprint object from a pickle dump.

        Parameters
        ----------
        path_or_bytes : str, pathlib.Path or bytes
            The path to the pickle file, or bytes corresponding to a pickle
            dump.


        .. seealso::

            :meth:`~Fingerprint.to_pickle` for creating the pickle dump.

        .. versionadded:: 1.0.0

        .. versionchanged:: 2.0.0
            Switched to dill instead of pickle

        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                r"The .+ interaction has been superseded by a new class",  # pragma: no cover
            )
            if isinstance(path_or_bytes, bytes):
                return dill.loads(path_or_bytes)
            with open(path_or_bytes, "rb") as f:
                return dill.load(f)

    def plot_lignetwork(
        self,
        ligand_mol,
        *,
        kind: Literal["aggregate", "frame"] = "aggregate",
        frame: int = 0,
        display_all: bool = False,
        threshold: float = 0.3,
        use_coordinates: bool = False,
        flatten_coordinates: bool = True,
        kekulize: bool = False,
        molsize: int = 35,
        rotation: float = 0,
        carbon: float = 0.16,
        width: str = "100%",
        height: str = "500px",
    ):
        """Generate and display a :class:`~prolif.plotting.network.LigNetwork` plot from
        a fingerprint object that has been used to run an analysis.

        Parameters
        ----------
        ligand_mol : rdkit.Chem.rdChem.Mol
            Ligand molecule.
        kind : str
            One of ``"aggregate"`` or ``"frame"``.
        frame : int
            Frame number (see :attr:`~prolif.fingerprint.Fingerprint.ifp`). Only
            applicable for ``kind="frame"``.
        display_all : bool
            Display all occurences for a given pair of residues and interaction, or only
            the shortest one. Only applicable for ``kind="frame"``. Not relevant if
            ``count=False`` in the ``Fingerprint`` object.
        threshold : float
            Frequency threshold, between 0 and 1. Only applicable for
            ``kind="aggregate"``.
        use_coordinates : bool
            If ``True``, uses the coordinates of the molecule directly, otherwise generates
            2D coordinates from scratch. See also ``flatten_coordinates``.
        flatten_coordinates : bool
            If this is ``True`` and ``use_coordinates=True``, generates 2D coordinates that
            are constrained to fit the 3D conformation of the ligand as best as possible.
        kekulize : bool
            Kekulize the ligand.
        molsize : int
            Multiply the coordinates by this number to create a bigger and
            more readable depiction.
        rotation : int
            Rotate the structure on the XY plane.
        carbon : float
            Size of the carbon atom dots on the depiction. Use `0` to hide the
            carbon dots.
        width : str
            Width of the IFrame window.
        height : str
            Height of the IFrame window.

        See Also
        --------
        :class:`prolif.plotting.network.LigNetwork`

        .. versionadded:: 2.0.0
        """
        from prolif.plotting.network import LigNetwork

        ligplot = LigNetwork.from_fingerprint(
            fp=self,
            ligand_mol=ligand_mol,
            kind=kind,
            frame=frame,
            display_all=display_all,
            threshold=threshold,
            use_coordinates=use_coordinates,
            flatten_coordinates=flatten_coordinates,
            kekulize=kekulize,
            molsize=molsize,
            rotation=rotation,
            carbon=carbon,
        )
        return ligplot.display(width=width, height=height)

    def plot_barcode(
        self,
        *,
        figsize: Tuple[int, int] = (8, 10),
        dpi: int = 100,
        interactive: bool = IS_NOTEBOOK,
        n_frame_ticks: int = 10,
        residues_tick_location: Literal["top", "bottom"] = "top",
        xlabel: str = "Frame",
        subplots_kwargs: Optional[dict] = None,
        tight_layout_kwargs: Optional[dict] = None,
    ):
        """Generate and display a :class:`~prolif.plotting.barcode.Barcode` plot from
        a fingerprint object that has been used to run an analysis.

        Parameters
        ----------
        figsize: Tuple[int, int] = (8, 10)
            Size of the matplotlib figure.
        dpi: int = 100
            DPI used for the matplotlib figure.
        interactive: bool
            Add hover interactivity to the plot (only relevant for notebooks). You may
            need to add ``%matplotlib notebook`` or ``%matplotlib ipympl`` for it to
            work as expected.
        n_frame_ticks: int = 10
            Number of ticks on the X axis. May use ±1 tick to have them evenly spaced.
        residues_tick_location: Literal["top", "bottom"] = "top"
            Whether the Y ticks appear at the top or at the bottom of the series of
            interactions of each residue.
        xlabel: str = "Frame"
            Label displayed for the X axis.
        subplots_kwargs: Optional[dict] = None
            Other parameters passed to :func:`matplotlib.pyplot.subplots`.
        tight_layout_kwargs: Optional[dict] = None
            Other parameters passed to :meth:`matplotlib.figure.Figure.tight_layout`.

        See Also
        --------
        :class:`prolif.plotting.barcode.Barcode`

        .. versionadded:: 2.0.0
        """
        from prolif.plotting.barcode import Barcode

        barcode = Barcode.from_fingerprint(self)
        return barcode.display(
            figsize=figsize,
            dpi=dpi,
            interactive=interactive,
            n_frame_ticks=n_frame_ticks,
            residues_tick_location=residues_tick_location,
            xlabel=xlabel,
            subplots_kwargs=subplots_kwargs,
            tight_layout_kwargs=tight_layout_kwargs,
        )

    def plot_3d(
        self,
        ligand_mol: Molecule,
        protein_mol: Molecule,
        *,
        frame: int,
        size: Tuple[int, int] = (650, 600),
        display_all: bool = False,
    ):
        """Generate and display the complex in 3D with py3Dmol from a fingerprint object
        that has been used to run an analysis.

        Parameters
        ----------
        ligand_mol : Molecule
            The ligand molecule to display.
        protein_mol : Molecule
            The protein molecule to display.
        frame : int
            The frame number chosen to select which interactions are going to be
            displayed.
        size: Tuple[int, int] = (650, 600)
            The size of the py3Dmol widget view.
        display_all : bool
            Display all occurences for a given pair of residues and interaction, or only
            the shortest one. Not relevant if ``count=False`` in the ``Fingerprint``
            object.

        See Also
        --------
        :class:`prolif.plotting.complex3d.Complex3D`

        .. versionadded:: 2.0.0
        """
        from prolif.plotting.complex3d import Complex3D

        plot3d = Complex3D.from_fingerprint(
            self, frame=frame, lig_mol=ligand_mol, prot_mol=protein_mol
        )
        return plot3d.display(size=size, display_all=display_all)



"""
Utility functions for interactions --- :mod:`prolif.interactions.utils`
=======================================================================

This module contains some utilities used by the interaction classes.
"""


def get_mapindex(res, index):
    """Get the index of the atom in the original molecule

    Parameters
    ----------
    res : prolif.residue.Residue
        The residue in the protein or ligand
    index : int
        The index of the atom in the :class:`~prolif.residue.Residue`

    Returns
    -------
    mapindex : int
        The index of the atom in the :class:`~prolif.molecule.Molecule`
    """
    return res.GetAtomWithIdx(index).GetUnsignedProp("mapindex")


def _distance_3args_l1_p1(l1, p1, p2):
    return l1.Distance(p1)


def _distance_3args_l1_p2(l1, p1, p2):
    return l1.Distance(p2)


def _distance_4args_l1_p1(l1, l2, p1, p2):
    return l1.Distance(p1)


def _distance_4args_l1_p2(l1, l2, p1, p2):
    return l1.Distance(p2)


def _distance_4args_l2_p1(l1, l2, p1, p2):
    return l2.Distance(p1)


def _distance_4args_l2_p2(l1, l2, p1, p2):
    return l2.Distance(p2)


DISTANCE_FUNCTIONS = {
    "P1": _distance_3args_l1_p1,
    "P2": _distance_3args_l1_p2,
    ("L1", "P1"): _distance_4args_l1_p1,
    ("L1", "P2"): _distance_4args_l1_p2,
    ("L2", "P1"): _distance_4args_l2_p1,
    ("L2", "P2"): _distance_4args_l2_p2,
}



"""
Detecting interactions between residues --- :mod:`prolif.interactions.interactions`
===================================================================================

Note that some of the SMARTS patterns used in the interaction classes are inspired from
`Pharmit`_ and `RDKit`_.

.. _Pharmit: https://sourceforge.net/p/pharmit/code/ci/master/tree/src/pharmarec.cpp
.. _RDKit: https://github.com/rdkit/rdkit/blob/master/Data/BaseFeatures.fdef

"""

from itertools import product
from math import degrees, radians

from MDAnalysis.topology.tables import vdwradii
from rdkit import Geometry
from rdkit.Chem import MolFromSmarts

from prolif.interactions.base import (
    BasePiStacking,
    Distance,
    DoubleAngle,
    Interaction,
    SingleAngle,
)
from prolif.utils import angle_between_limits, get_centroid, get_ring_normal_vector

__all__ = [
    "Hydrophobic",
    "HBAcceptor",
    "HBDonor",
    "XBAcceptor",
    "XBDonor",
    "Cationic",
    "Anionic",
    "CationPi",
    "PiCation",
    "FaceToFace",
    "EdgeToFace",
    "PiStacking",
    "MetalDonor",
    "MetalAcceptor",
    "VdWContact",
]
VDWRADII = {symbol.capitalize(): radius for symbol, radius in vdwradii.items()}


class Hydrophobic(Distance):
    """Hydrophobic interaction

    Parameters
    ----------
    hydrophobic : str
        SMARTS query for hydrophobic atoms
    distance : float
        Distance threshold for the interaction


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    """

    def __init__(
        self,
        hydrophobic=(
            "[c,s,Br,I,S&H0&v2," "$([D3,D4;#6])&!$([#6]~[#7,#8,#9])&!$([#6X4H0]);+0]"
        ),
        distance=4.5,
    ):
        super().__init__(
            lig_pattern=hydrophobic, prot_pattern=hydrophobic, distance=distance
        )


class HBAcceptor(SingleAngle):
    """Hbond interaction between a ligand (acceptor) and a residue (donor)

    Parameters
    ----------
    acceptor : str
        SMARTS for ``[Acceptor]``
    donor : str
        SMARTS for ``[Donor]-[Hydrogen]``
    distance : float
        Distance threshold between the acceptor and donor atoms
    DHA_angle : tuple
        Min and max values for the ``[Acceptor]...[Hydrogen]-[Donor]`` angle


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    .. versionchanged:: 2.0.0
        ``angles`` parameter renamed to ``DHA_angle``.

    """

    def __init__(
        self,
        acceptor=(
            "[#7&!$([nX3])&!$([NX3]-*=[O,N,P,S])&!$([NX3]-[a])&!$([Nv4&+1]),"
            "O&!$([OX2](C)C=O)&!$(O(~a)~a)&!$(O=N-*)&!$([O-]-N=O),o+0,"
            "F&$(F-[#6])&!$(F-[#6][F,Cl,Br,I])]"
        ),
        donor="[$([O,S;+0]),$([N;v3,v4&+1]),n+0]-[H]",
        distance=3.5,
        DHA_angle=(130, 180),
    ):
        super().__init__(
            lig_pattern=acceptor,
            prot_pattern=donor,
            distance=distance,
            angle=DHA_angle,
            distance_atom="P1",
            metadata_mapping={"angle": "DHA_angle"},
        )


HBDonor = HBAcceptor.invert_role(
    "HBDonor", "Hbond interaction between a ligand (donor) and a residue (acceptor)"
)


class XBAcceptor(DoubleAngle):
    """Halogen bonding between a ligand (acceptor) and a residue (donor).

    Parameters
    ----------
    acceptor : str
        SMARTS for ``[Acceptor]-[R]``
    donor : str
        SMARTS for ``[Donor]-[Halogen]``
    distance : float
        Cutoff distance between the acceptor and halogen atoms
    AXD_angle : tuple
        Min and max values for the ``[Acceptor]...[Halogen]-[Donor]`` angle
    XAR_angle : tuple
        Min and max values for the ``[Halogen]...[Acceptor]-[R]`` angle

    Notes
    -----
    Distance and angle adapted from Auffinger et al. PNAS 2004


    .. versionchanged:: 2.0.0
        ``axd_angles`` and ``xar_angles`` parameters renamed to ``AXD_angle`` and
        ``XAR_angle``.

    """

    def __init__(
        self,
        acceptor="[#7,#8,P,S,Se,Te,a;!+{1-}][*]",
        donor="[#6,#7,Si,F,Cl,Br,I]-[Cl,Br,I,At]",
        distance=3.5,
        AXD_angle=(130, 180),
        XAR_angle=(80, 140),
    ):
        super().__init__(
            lig_pattern=acceptor,
            prot_pattern=donor,
            distance=distance,
            L1P2P1_angle=AXD_angle,
            L2L1P2_angle=XAR_angle,
            distance_atoms=("L1", "P2"),
            metadata_mapping={
                "L1P2P1_angle": "AXD_angle",
                "L2L1P2_angle": "XAR_angle",
            },
        )


XBDonor = XBAcceptor.invert_role(
    "XBDonor", "Halogen bonding between a ligand (donor) and a residue (acceptor)"
)


class Cationic(Distance):
    """Ionic interaction between a ligand (cation) and a residue (anion)

    .. versionchanged:: 1.1.0
        Handles resonance forms for common acids, amidine and guanidine.

    """

    def __init__(
        self,
        cation="[+{1-},$([NX3&!$([NX3]-O)]-[C]=[NX3+])]",
        anion="[-{1-},$(O=[C,S,P]-[O-])]",
        distance=4.5,
    ):
        super().__init__(lig_pattern=cation, prot_pattern=anion, distance=distance)


Anionic = Cationic.invert_role(
    "Anionic", "Ionic interaction between a ligand (anion) and a residue (cation)"
)


class CationPi(Interaction):
    """Cation-Pi interaction between a ligand (cation) and a residue (aromatic ring)

    Parameters
    ----------
    cation : str
        SMARTS for cation
    pi_ring : tuple
        SMARTS for aromatic rings (5 and 6 membered rings only)
    distance : float
        Cutoff distance between the centroid and the cation
    angle : tuple
        Min and max values for the angle between the vector normal to the ring
        plane and the vector going from the centroid to the cation


    .. versionchanged:: 1.1.0
        Handles resonance forms for amidine and guanidine as cations.

    .. versionchanged:: 2.0.0
        ``angles`` parameter renamed to ``angle``.

    """

    def __init__(
        self,
        cation="[+{1-},$([NX3&!$([NX3]-O)]-[C]=[NX3+])]",
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
        distance=4.5,
        angle=(0, 30),
    ):
        self.cation = MolFromSmarts(cation)
        self.pi_ring = [MolFromSmarts(s) for s in pi_ring]
        self.distance = distance
        self.angle = tuple(radians(i) for i in angle)

    def detect(self, cation, pi):
        cation_matches = cation.GetSubstructMatches(self.cation)
        for pi_ring in self.pi_ring:
            pi_matches = pi.GetSubstructMatches(pi_ring)
            if not (cation_matches and pi_matches):
                continue
            for cation_match, pi_match in product(cation_matches, pi_matches):
                cat = Geometry.Point3D(*cation.xyz[cation_match[0]])
                # get coordinates of atoms matching pi-system
                pi_coords = pi.xyz[list(pi_match)]
                # centroid of pi-system as 3d point
                centroid = Geometry.Point3D(*get_centroid(pi_coords))
                # distance between cation and centroid
                dist = cat.Distance(centroid)
                if dist > self.distance:
                    continue
                # vector normal to ring plane
                normal = get_ring_normal_vector(centroid, pi_coords)
                # vector between the centroid and the charge
                centroid_cation = centroid.DirectionVector(cat)
                # compute angle between normal to ring plane and
                # centroid-cation
                angle = normal.AngleTo(centroid_cation)
                if angle_between_limits(angle, *self.angle, ring=True):
                    yield self.metadata(
                        cation,
                        pi,
                        cation_match,
                        pi_match,
                        distance=dist,
                        angle=degrees(angle),
                    )


PiCation = CationPi.invert_role(
    "PiCation",
    "Cation-Pi interaction between a ligand (aromatic ring) and a residue (cation)",
)


class FaceToFace(BasePiStacking):
    """Face-to-face Pi-Stacking interaction between a ligand and a residue

    Parameters
    ----------
    distance : float
        Cutoff distance between each rings centroid
    plane_angles : tuple
        Min and max values for the angle between the ring planes
    normal_centroids_angles : tuple
        Min and max angles allowed between the vector normal to a ring's plane,
        and the vector between the centroid of both rings.
    pi_ring : list
        List of SMARTS for aromatic rings


    .. versionchanged:: 2.0.0
        Renamed ``centroid_distance`` to distance.

    """

    def __init__(
        self,
        distance=5.5,
        plane_angle=(0, 35),
        normal_to_centroid_angle=(0, 33),
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
    ):
        super().__init__(
            distance=distance,
            plane_angle=plane_angle,
            normal_to_centroid_angle=normal_to_centroid_angle,
            pi_ring=pi_ring,
        )


class EdgeToFace(BasePiStacking):
    """Edge-to-face Pi-Stacking interaction between a ligand and a residue

    Parameters
    ----------
    distance : float
        Cutoff distance between each rings centroid
    plane_angles : tuple
        Min and max values for the angle between the ring planes
    normal_centroids_angles : tuple
        Min and max angles allowed between the vector normal to a ring's plane,
        and the vector between the centroid of both rings.
    pi_ring : list
        List of SMARTS for aromatic rings
    intersect_radius : float
        Used to check whether the intersect point between ring planes falls within
        ``intersect_radius`` of the opposite ring's centroid.


    .. versionchanged:: 1.1.0
        In addition to the changes made to the base pi-stacking interaction, this
        implementation makes sure that the intersection between the perpendicular ring's
        plane and the other's plane falls inside the ring.

    .. versionchanged:: 2.0.0
        Renamed ``centroid_distance`` to distance, added ``intersect_radius`` parameter.

    """

    def __init__(
        self,
        distance=6.5,
        plane_angle=(50, 90),
        normal_to_centroid_angle=(0, 30),
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
        intersect_radius=1.5,
    ):
        super().__init__(
            distance=distance,
            plane_angle=plane_angle,
            normal_to_centroid_angle=normal_to_centroid_angle,
            pi_ring=pi_ring,
            intersect=True,
            intersect_radius=intersect_radius,
        )


class PiStacking(Interaction):
    """Pi-Stacking interaction between a ligand and a residue

    Parameters
    ----------
    ftf_kwargs : dict
        Parameters to pass to the underlying FaceToFace class
    etf_kwargs : dict
        Parameters to pass to the underlying EdgeToFace class


    .. versionchanged:: 0.3.4
        `shortest_distance` has been replaced by `angle_normal_centroid`

    .. versionchanged:: 1.1.0
        The implementation now directly calls :class:`EdgeToFace` and :class:`FaceToFace`
        instead of overwriting the default parameters with more generic ones.

    """

    def __init__(self, ftf_kwargs=None, etf_kwargs=None):
        self.ftf = FaceToFace(**ftf_kwargs or {})
        self.etf = EdgeToFace(**etf_kwargs or {})

    def detect(self, ligand, residue):
        yield from self.ftf.detect(ligand, residue)
        yield from self.etf.detect(ligand, residue)


class MetalDonor(Distance):
    """Metal complexation interaction between a ligand (metal) and a residue (chelated)

    Parameters
    ----------
    metal : str
        SMARTS for a transition metal
    ligand : str
        SMARTS for a ligand
    distance : float
        Cutoff distance


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    """

    def __init__(
        self,
        metal="[Ca,Cd,Co,Cu,Fe,Mg,Mn,Ni,Zn]",
        ligand="[O,#7&!$([nX3])&!$([NX3]-*=[!#6])&!$([NX3]-[a])&!$([NX4]),-{1-};!+{1-}]",
        distance=2.8,
    ):
        super().__init__(lig_pattern=metal, prot_pattern=ligand, distance=distance)


MetalAcceptor = MetalDonor.invert_role(
    "MetalAcceptor",
    "Metal complexation interaction between a ligand (chelated) and a residue (metal)",
)


class VdWContact(Interaction):
    """Interaction based on the van der Waals radii of interacting atoms.

    Parameters
    ----------
    tolerance : float
        Tolerance added to the sum of vdW radii of atoms before comparing to
        the interatomic distance. If ``distance <= sum_vdw + tolerance`` the
        atoms are identified as a contact
    vdwradii : dict, optional
        Updates to the vdW radii dictionary, with elements (first letter uppercase) as a
        key and the radius as a value.

    Raises
    ------
    ValueError
        ``tolerance`` parameter cannot be negative


    .. versionchanged:: 2.0.0
        Added the ``vdwradii`` parameter.

    """

    def __init__(self, tolerance=0.0, vdwradii=None):
        if tolerance >= 0:
            self.tolerance = tolerance
        else:
            raise ValueError("`tolerance` must be 0 or positive")
        self._vdw_cache = {}
        self.vdwradii = {**VDWRADII, **vdwradii} if vdwradii else VDWRADII

    def detect(self, ligand, residue):
        lxyz = ligand.GetConformer()
        rxyz = residue.GetConformer()
        for la, ra in product(ligand.GetAtoms(), residue.GetAtoms()):
            lig = la.GetSymbol()
            res = ra.GetSymbol()
            elements = frozenset((lig, res))
            try:
                vdw = self._vdw_cache[elements]
            except KeyError:
                vdw = self.vdwradii[lig] + self.vdwradii[res] + self.tolerance
                self._vdw_cache[elements] = vdw
            dist = lxyz.GetAtomPosition(la.GetIdx()).Distance(
                rxyz.GetAtomPosition(ra.GetIdx())
            )
            if dist <= vdw:
                yield self.metadata(
                    ligand, residue, (la.GetIdx(),), (ra.GetIdx(),), distance=dist
                )



"""
Plot interactions in 3D --- :mod:`prolif.plotting.complex3d`
============================================================

.. versionadded:: 2.0.0

.. autoclass:: Complex3D
   :members:

"""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, Set, Tuple

import py3Dmol
from rdkit import Chem
from rdkit.Geometry import Point3D

from prolif.exceptions import RunRequiredError
from prolif.plotting.utils import separated_interaction_colors
from prolif.utils import get_centroid

if TYPE_CHECKING:
    from prolif.fingerprint import Fingerprint
    from prolif.ifp import IFP
    from prolif.molecule import Molecule


class Complex3D:
    """Creates a py3Dmol plot of interactions.

    Parameters
    ----------
    ifp : IFP
        The interaction dictionary for a single frame.
    lig_mol : Molecule
        The ligand molecule to display.
    prot_mol : Molecule
        The protein molecule to display.

    Attributes
    ----------
    COLORS : dict
        Dictionnary of colors used in the plot for interactions.
    LIGAND_STYLE : Dict[str, Dict] = {"stick": {"colorscheme": "cyanCarbon"}}
        Style object passed to ``3Dmol.js`` for the ligand.
    RESIDUES_STYLE : Dict[str, Dict] = {"stick": {}}
        Style object passed to ``3Dmol.js`` for the protein residues involved in
        interactions.
    PROTEIN_STYLE : Dict[str, Dict] = {"cartoon": {"style": "edged"}}
        Style object passed to ``3Dmol.js`` for the entire protein.
    PEPTIDE_STYLE : Dict[str, Dict] = "cartoon": {"style": "edged", "colorscheme": "cyanCarbon"}
        Style object passed to ``3Dmol.js`` for the ligand as a peptide if appropriate.
    PEPTIDE_THRESHOLD : int = 2
        Ligands with this number of residues or more will be displayed using
        ``PEPTIDE_STYLE`` in addition to the ``LIGAND_STYLE``.
    LIGAND_DISPLAYED_ATOM : Dict[str, int]
        Which atom should be used to display an atom-to-atom interaction for the ligand.
        Refers to the order defined in the SMARTS pattern used in interaction
        definition. Interactions not specified here use ``0`` by default.
    PROTEIN_DISPLAYED_ATOM : Dict[str, int]
        Same as :attr:`LIGAND_DISPLAYED_ATOM` for the protein.
    LIGAND_RING_INTERACTIONS : Set[str]
        Which interactions should be displayed using the centroid instead of using
        :attr:`LIGAND_DISPLAYED_ATOM` for the ligand.
    PROTEIN_RING_INTERACTIONS : Set[str]
        Which interactions should be displayed using the centroid instead of using
        :attr:`PROTEIN_DISPLAYED_ATOM` for the protein.
    RESIDUE_HOVER_CALLBACK : str
        JavaScript callback executed when hovering a residue involved in an interaction.
    INTERACTION_HOVER_CALLBACK : str
        JavaScript callback executed when hovering an interaction line.
    DISABLE_HOVER_CALLBACK : str
        JavaScript callback executed when the hovering event is finished.
    """

    COLORS: ClassVar[Dict[str, str]] = {**separated_interaction_colors}
    LIGAND_STYLE: ClassVar[Dict] = {"stick": {"colorscheme": "cyanCarbon"}}
    RESIDUES_STYLE: ClassVar[Dict] = {"stick": {}}
    PROTEIN_STYLE: ClassVar[Dict] = {"cartoon": {"style": "edged"}}
    PEPTIDE_STYLE: ClassVar[Dict] = {
        "cartoon": {"style": "edged", "colorscheme": "cyanCarbon"}
    }
    PEPTIDE_THRESHOLD: ClassVar[int] = 5
    LIGAND_DISPLAYED_ATOM = {
        "HBDonor": 1,
        "XBDonor": 1,
    }
    PROTEIN_DISPLAYED_ATOM = {
        "HBAcceptor": 1,
        "XBAcceptor": 1,
    }
    RING_SYSTEMS: ClassVar[Set[str]] = {
        "PiStacking",
        "EdgeToFace",
        "FaceToFace",
    }
    LIGAND_RING_INTERACTIONS: ClassVar[Set[str]] = {*RING_SYSTEMS, "PiCation"}
    PROTEIN_RING_INTERACTIONS: ClassVar[Set[str]] = {*RING_SYSTEMS, "CationPi"}
    RESIDUE_HOVER_CALLBACK: ClassVar[
        str
    ] = """function(atom,viewer) {
        if(!atom.label) {
            atom.label = viewer.addLabel('%s:'+atom.atom+atom.serial,
                {position: atom, backgroundColor: 'mintcream', fontColor:'black'});
        }
    }"""
    INTERACTION_HOVER_CALLBACK: ClassVar[
        str
    ] = """
    function(atom,viewer) {
        if(!atom.label) {
            atom.label = viewer.addLabel(atom.interaction,
                {position: atom, backgroundColor: 'black', fontColor:'white'});
        }
    }"""
    DISABLE_HOVER_CALLBACK: ClassVar[
        str
    ] = """
    function(atom,viewer) { 
        if(atom.label) {
            viewer.removeLabel(atom.label);
            delete atom.label;
        }
    }"""

    def __init__(self, ifp: IFP, lig_mol: Molecule, prot_mol: Molecule) -> None:
        self.ifp = ifp
        self.lig_mol = lig_mol
        self.prot_mol = prot_mol

    @classmethod
    def from_fingerprint(
        cls,
        fp: Fingerprint,
        lig_mol: Molecule,
        prot_mol: Molecule,
        *,
        frame: int,
    ) -> Complex3D:
        """Creates a py3Dmol plot of interactions.

        Parameters
        ----------
        fp : prolif.fingerprint.Fingerprint
            The fingerprint object already executed using one of the ``run`` or
            ``run_from_iterable`` methods.
        frame : int
            The frame number chosen to select which interactions are going to be
            displayed.
        lig_mol : Molecule
            The ligand molecule to display.
        prot_mol : Molecule
            The protein molecule to display.
        """
        if not hasattr(fp, "ifp"):
            raise RunRequiredError(
                "Please run the fingerprint analysis before attempting to display results."
            )
        ifp = fp.ifp[frame]
        return cls(ifp, lig_mol, prot_mol)

    @staticmethod
    def get_ring_centroid(mol: Molecule, indices: Tuple[int, ...]) -> Point3D:
        centroid = mol.xyz[list(indices)].mean(axis=0)
        return Point3D(*centroid)

    def display(self, size: Tuple[int, int] = (650, 600), display_all: bool = False):
        """Display as a py3Dmol widget view.

        Parameters
        ----------
        size: Tuple[int, int] = (650, 600)
            The size of the py3Dmol widget view.
        display_all : bool
            Display all occurences for a given pair of residues and interaction, or only
            the shortest one. Not relevant if ``count=False`` in the ``Fingerprint``
            object.
        """
        v = py3Dmol.view(width=size[0], height=size[1])
        v.removeAllModels()

        models = {}
        mid = -1
        for (lresid, presid), interactions in self.ifp.items():
            lres = self.lig_mol[lresid]
            pres = self.prot_mol[presid]
            # set model ids for reusing later
            for resid, res, style in [
                (lresid, lres, self.LIGAND_STYLE),
                (presid, pres, self.RESIDUES_STYLE),
            ]:
                if resid not in models:
                    mid += 1
                    v.addModel(Chem.MolToMolBlock(res), "sdf")
                    model = v.getModel()
                    model.setStyle({}, style)
                    # add residue label
                    model.setHoverable(
                        {},
                        True,
                        self.RESIDUE_HOVER_CALLBACK % resid,
                        self.DISABLE_HOVER_CALLBACK,
                    )
                    models[resid] = mid
            for interaction, metadata_tuple in interactions.items():
                # whether to display all interactions or only the one with the shortest
                # distance
                metadata_iterator = (
                    metadata_tuple
                    if display_all
                    else (
                        min(
                            metadata_tuple,
                            key=lambda m: m.get("distance", float("nan")),
                        ),
                    )
                )
                for metadata in metadata_iterator:
                    # get coordinates for both points of the interaction
                    if interaction in self.LIGAND_RING_INTERACTIONS:
                        p1 = self.get_ring_centroid(lres, metadata["indices"]["ligand"])
                    else:
                        p1 = lres.GetConformer().GetAtomPosition(
                            metadata["indices"]["ligand"][
                                self.LIGAND_DISPLAYED_ATOM.get(interaction, 0)
                            ]
                        )
                    if interaction in self.PROTEIN_RING_INTERACTIONS:
                        p2 = self.get_ring_centroid(
                            pres, metadata["indices"]["protein"]
                        )
                    else:
                        p2 = pres.GetConformer().GetAtomPosition(
                            metadata["indices"]["protein"][
                                self.PROTEIN_DISPLAYED_ATOM.get(interaction, 0)
                            ]
                        )
                    # add interaction line
                    v.addCylinder(
                        {
                            "start": dict(x=p1.x, y=p1.y, z=p1.z),
                            "end": dict(x=p2.x, y=p2.y, z=p2.z),
                            "color": self.COLORS.get(interaction, "grey"),
                            "radius": 0.15,
                            "dashed": True,
                            "fromCap": 1,
                            "toCap": 1,
                        }
                    )
                    # add label when hovering the middle of the dashed line by adding a dummy atom
                    c = Point3D(*get_centroid([p1, p2]))
                    modelID = models[lresid]
                    model = v.getModel(modelID)
                    interaction_label = f"{interaction}: {metadata['distance']:.2f}Å"
                    model.addAtoms(
                        [
                            {
                                "elem": "Z",
                                "x": c.x,
                                "y": c.y,
                                "z": c.z,
                                "interaction": interaction_label,
                            }
                        ]
                    )
                    model.setStyle(
                        {"interaction": interaction_label},
                        {"clicksphere": {"radius": 0.5}},
                    )
                    model.setHoverable(
                        {"interaction": interaction_label},
                        True,
                        self.INTERACTION_HOVER_CALLBACK,
                        self.DISABLE_HOVER_CALLBACK,
                    )

        # show protein
        mol = Chem.RemoveAllHs(self.prot_mol)
        pdb = Chem.MolToPDBBlock(mol, flavor=0x20 | 0x10)
        v.addModel(pdb, "pdb")
        model = v.getModel()
        model.setStyle({}, self.PROTEIN_STYLE)

        # do the same for ligand if multiple residues
        if self.lig_mol.n_residues >= self.PEPTIDE_THRESHOLD:
            mol = Chem.RemoveAllHs(self.lig_mol)
            pdb = Chem.MolToPDBBlock(mol, flavor=0x20 | 0x10)
            v.addModel(pdb, "pdb")
            model = v.getModel()
            model.setStyle({}, self.PEPTIDE_STYLE)

        v.zoomTo({"model": list(models.values())})
        return v
